import asyncio
import os
from unittest.mock import ANY, AsyncMock, MagicMock, patch

import pytest

from aiocache.base import API, BaseCache, _Conn, _ensure_key
from ..utils import Keys


class TestAPI:
    def test_register(self):
        @API.register
        def dummy():
            """Dummy function."""

        assert dummy in API.CMDS
        API.unregister(dummy)

    def test_unregister(self):
        @API.register
        def dummy():
            "Dummy function."""

        API.unregister(dummy)
        assert dummy not in API.CMDS

    def test_unregister_unexisting(self):
        def dummy():
            "Dummy function."""

        API.unregister(dummy)
        assert dummy not in API.CMDS

    async def test_aiocache_enabled(self):
        @API.aiocache_enabled()
        async def dummy(*args, **kwargs):
            return True

        assert await dummy() is True

    async def test_aiocache_enabled_disabled(self):
        @API.aiocache_enabled(fake_return=[])
        async def dummy(*args, **kwargs):
            """Dummy function."""

        with patch.dict(os.environ, {"AIOCACHE_DISABLE": "1"}):
            assert await dummy() == []

    async def test_timeout_no_timeout(self):
        self = MagicMock(spec_set=("timeout",))
        self.timeout = 0

        @API.timeout
        async def dummy(self):
            self()

        with patch("asyncio.wait_for") as wait_for:
            await dummy(self)
            assert self.call_count == 1
            assert wait_for.call_count == 0

    async def test_timeout_self(self):
        self = MagicMock(spec_set=("timeout",))
        self.timeout = 0.002

        @API.timeout
        async def dummy(self):
            await asyncio.sleep(0.005)

        with pytest.raises(asyncio.TimeoutError):
            await dummy(self)

    async def test_timeout_kwarg_0(self):
        self = MagicMock(spec_set=("timeout",))
        self.timeout = 0.002

        @API.timeout
        async def dummy(self):
            await asyncio.sleep(0.005)
            return True

        assert await dummy(self, timeout=0) is True

    async def test_timeout_kwarg_None(self):
        self = MagicMock(spec_set=("timeout",))
        self.timeout = 0.002

        @API.timeout
        async def dummy(self):
            await asyncio.sleep(0.005)
            return True

        assert await dummy(self, timeout=None) is True

    async def test_timeout_kwarg(self):
        self = MagicMock(spec_set=("timeout",))

        @API.timeout
        async def dummy(self):
            await asyncio.sleep(0.005)

        with pytest.raises(asyncio.TimeoutError):
            await dummy(self, timeout=0.002)

    async def test_timeout_self_kwarg(self):
        self = MagicMock(spec_set=("timeout",))
        self.timeout = 5

        @API.timeout
        async def dummy(self):
            await asyncio.sleep(0.005)

        with pytest.raises(asyncio.TimeoutError):
            await dummy(self, timeout=0.003)

    async def test_plugins(self):
        self = MagicMock(spec_set=("plugins",))
        plugin1 = MagicMock(spec_set=("pre_dummy", "post_dummy"))
        plugin1.pre_dummy = AsyncMock(spec_set=())
        plugin1.post_dummy = AsyncMock(spec_set=())
        plugin2 = MagicMock(spec_set=("pre_dummy", "post_dummy"))
        plugin2.pre_dummy = AsyncMock(spec_set=())
        plugin2.post_dummy = AsyncMock(spec_set=())
        self.plugins = (plugin1, plugin2)

        @API.plugins
        async def dummy(self, *args, **kwargs):
            return True

        assert await dummy(self) is True
        plugin1.pre_dummy.assert_called_with(self)
        plugin1.post_dummy.assert_called_with(self, took=ANY, ret=True)
        plugin2.pre_dummy.assert_called_with(self)
        plugin2.post_dummy.assert_called_with(self, took=ANY, ret=True)


class TestBaseCache:
    def test_str_ttl(self):
        cache = BaseCache(ttl="1.5")
        assert cache.ttl == 1.5

    def test_str_timeout(self):
        cache = BaseCache(timeout="1.5")
        assert cache.timeout == 1.5

    async def test_add(self, base_cache):
        with pytest.raises(NotImplementedError):
            await base_cache._add(Keys.KEY, "value", 0)

    async def test_get(self, base_cache):
        with pytest.raises(NotImplementedError):
            await base_cache._get(Keys.KEY, "utf-8")

    async def test_set(self, base_cache):
        with pytest.raises(NotImplementedError):
            await base_cache._set(Keys.KEY, "value", 0)

    async def test_multi_get(self, base_cache):
        with pytest.raises(NotImplementedError):
            await base_cache._multi_get([Keys.KEY], encoding="utf-8")

    async def test_multi_set(self, base_cache):
        with pytest.raises(NotImplementedError):
            await base_cache._multi_set([(Keys.KEY, "value")], 0)

    async def test_delete(self, base_cache):
        with pytest.raises(NotImplementedError):
            await base_cache._delete(Keys.KEY)

    async def test_exists(self, base_cache):
        with pytest.raises(NotImplementedError):
            await base_cache._exists(Keys.KEY)

    async def test_increment(self, base_cache):
        with pytest.raises(NotImplementedError):
            await base_cache._increment(Keys.KEY, 2)

    async def test_expire(self, base_cache):
        with pytest.raises(NotImplementedError):
            await base_cache._expire(Keys.KEY, 0)

    async def test_clear(self, base_cache):
        with pytest.raises(NotImplementedError):
            await base_cache._clear("namespace")

    async def test_raw(self, base_cache):
        with pytest.raises(NotImplementedError):
            await base_cache._raw("get", Keys.KEY)

    async def test_close(self, base_cache):
        assert await base_cache._close() is None

    async def test_acquire_conn(self, base_cache):
        assert await base_cache.acquire_conn() == base_cache

    async def test_release_conn(self, base_cache):
        assert await base_cache.release_conn("mock") is None

    @pytest.fixture
    def set_test_namespace(self, base_cache):
        base_cache.namespace = "test"
        yield
        base_cache.namespace = None

    @pytest.mark.parametrize(
        "namespace, expected",
        ([None, "test" + _ensure_key(Keys.KEY)], ["", _ensure_key(Keys.KEY)], ["my_ns", "my_ns" + _ensure_key(Keys.KEY)]),  # type: ignore[attr-defined]  # noqa: B950
    )
    def test_build_key(self, set_test_namespace, base_cache, namespace, expected):
        assert base_cache.build_key(Keys.KEY, namespace=namespace) == expected

    def test_alt_build_key(self):
        cache = BaseCache(key_builder=lambda key, namespace: "x")
        assert cache.build_key(Keys.KEY, "namespace") == "x"

    @pytest.fixture
    def alt_base_cache(self, init_namespace="test"):
        """Custom key_builder for cache"""
        def build_key(key, namespace=None):
            ns = namespace if namespace is not None else ""
            sep = ":" if namespace else ""
            return f"{ns}{sep}{_ensure_key(key)}"

        cache = BaseCache(key_builder=build_key, namespace=init_namespace)
        return cache

    @pytest.mark.parametrize(
        "namespace, expected",
        ([None, _ensure_key(Keys.KEY)], ["", _ensure_key(Keys.KEY)], ["my_ns", "my_ns:" + _ensure_key(Keys.KEY)]),  # type: ignore[attr-defined]  # noqa: B950
    )
    def test_alt_build_key_override_namespace(self, alt_base_cache, namespace, expected):
        """Custom key_builder overrides namespace of cache"""
        cache = alt_base_cache
        assert cache.build_key(Keys.KEY, namespace=namespace) == expected

    @pytest.mark.parametrize(
        "init_namespace, expected",
        ([None, _ensure_key(Keys.KEY)], ["", _ensure_key(Keys.KEY)], ["test", "test:" + _ensure_key(Keys.KEY)]),  # type: ignore[attr-defined]  # noqa: B950
    )
    async def test_alt_build_key_default_namespace(
            self, init_namespace, alt_base_cache, expected):
        """Custom key_builder for cache with or without namespace specified.

        Cache member functions that accept a ``namespace`` parameter
        should default to using ``self.namespace`` if the ``namespace``
        argument is ``None``.

        This enables a cache to correctly build keys when the cache is
        initialized with both a ``namespace`` and a ``key_builder``,
        even when that cache is supplied to a lock or to a decorator
        using the ``alias`` argument.
        """
        cache = alt_base_cache
        cache.namespace = init_namespace

        # Verify that private members are called with the correct ns_key
        await self._assert_add__alt_build_key_default_namespace(cache, expected)
        await self._assert_get__alt_build_key_default_namespace(cache, expected)
        await self._assert_multi_get__alt_build_key_default_namespace(cache, expected)
        await self._assert_set__alt_build_key_default_namespace(cache, expected)
        await self._assert_multi_set__alt_build_key_default_namespace(cache, expected)
        await self._assert_exists__alt_build_key_default_namespace(cache, expected)
        await self._assert_increment__alt_build_key_default_namespace(cache, expected)
        await self._assert_delete__alt_build_key_default_namespace(cache, expected)
        await self._assert_expire__alt_build_key_default_namespace(cache, expected)

    async def _assert_add__alt_build_key_default_namespace(self, cache, expected):
        with patch.object(cache, "_add", autospec=True) as _add:
            await cache.add(Keys.KEY, "value")
            _add.assert_called_once_with(expected, "value", _conn=None, ttl=None)

    async def _assert_get__alt_build_key_default_namespace(self, cache, expected):
        with patch.object(cache, "_get", autospec=True) as _get:
            await cache.get(Keys.KEY)
            _get.assert_called_once_with(
                expected, _conn=None, encoding=cache.serializer.encoding)

    async def _assert_multi_get__alt_build_key_default_namespace(self, cache, expected):
        with patch.object(cache, "_multi_get", autospec=True) as _multi_get:
            await cache.multi_get([Keys.KEY])
            _multi_get.assert_called_once_with(
                [expected], _conn=None, encoding=cache.serializer.encoding)

    async def _assert_set__alt_build_key_default_namespace(self, cache, expected):
        with patch.object(cache, "_set", autospec=True) as _set:
            await cache.set(Keys.KEY, "value")
            _set.assert_called_once_with(
                expected, "value", _conn=None, ttl=None, _cas_token=None)

    async def _assert_multi_set__alt_build_key_default_namespace(self, cache, expected):
        with patch.object(cache, "_multi_set", autospec=True) as _multi_set:
            await cache.multi_set([(Keys.KEY, "value")])
            _multi_set.assert_called_once_with(
                [(expected, "value")], _conn=None, ttl=None)

    async def _assert_exists__alt_build_key_default_namespace(self, cache, expected):
        with patch.object(cache, "_exists", autospec=True) as _exists:
            await cache.exists(Keys.KEY)
            _exists.assert_called_once_with(expected, _conn=None)

    async def _assert_increment__alt_build_key_default_namespace(self, cache, expected):
        with patch.object(cache, "_increment", autospec=True) as _increment:
            await cache.increment(Keys.KEY)
            _increment.assert_called_once_with(expected, delta=1, _conn=None)

    async def _assert_delete__alt_build_key_default_namespace(self, cache, expected):
        with patch.object(cache, "_delete", autospec=True) as _delete:
            await cache.delete(Keys.KEY)
            _delete.assert_called_once_with(expected, _conn=None)

    async def _assert_expire__alt_build_key_default_namespace(self, cache, expected):
        with patch.object(cache, "_expire", autospec=True) as _expire:
            await cache.expire(Keys.KEY, 0)
            _expire.assert_called_once_with(expected, 0, _conn=None)

    async def test_add_ttl_cache_default(self, base_cache):
        with patch.object(base_cache, "_add", autospec=True) as m:
            await base_cache.add(Keys.KEY, "value")

            m.assert_called_once_with(Keys.KEY, "value", _conn=None, ttl=None)

    async def test_add_ttl_default(self, base_cache):
        base_cache.ttl = 10
        with patch.object(base_cache, "_add", autospec=True) as m:
            await base_cache.add(Keys.KEY, "value")

            m.assert_called_once_with(Keys.KEY, "value", _conn=None, ttl=10)

    async def test_add_ttl_overriden(self, base_cache):
        base_cache.ttl = 10
        with patch.object(base_cache, "_add", autospec=True) as m:
            await base_cache.add(Keys.KEY, "value", ttl=20)

            m.assert_called_once_with(Keys.KEY, "value", _conn=None, ttl=20)

    async def test_add_ttl_none(self, base_cache):
        base_cache.ttl = 10
        with patch.object(base_cache, "_add", autospec=True) as m:
            await base_cache.add(Keys.KEY, "value", ttl=None)

            m.assert_called_once_with(Keys.KEY, "value", _conn=None, ttl=None)

    async def test_set_ttl_cache_default(self, base_cache):
        with patch.object(base_cache, "_set", autospec=True) as m:
            await base_cache.set(Keys.KEY, "value")

            m.assert_called_once_with(
                Keys.KEY, "value", _cas_token=None, _conn=None, ttl=None
            )

    async def test_set_ttl_default(self, base_cache):
        base_cache.ttl = 10
        with patch.object(base_cache, "_set", autospec=True) as m:
            await base_cache.set(Keys.KEY, "value")

            m.assert_called_once_with(
                Keys.KEY, "value", _cas_token=None, _conn=None, ttl=10
            )

    async def test_set_ttl_overriden(self, base_cache):
        base_cache.ttl = 10
        with patch.object(base_cache, "_set", autospec=True) as m:
            await base_cache.set(Keys.KEY, "value", ttl=20)

            m.assert_called_once_with(
                Keys.KEY, "value", _cas_token=None, _conn=None, ttl=20
            )

    async def test_set_ttl_none(self, base_cache):
        base_cache.ttl = 10
        with patch.object(base_cache, "_set", autospec=True) as m:
            await base_cache.set(Keys.KEY, "value", ttl=None)

            m.assert_called_once_with(
                Keys.KEY, "value", _cas_token=None, _conn=None, ttl=None
            )

    async def test_multi_set_ttl_cache_default(self, base_cache):
        with patch.object(base_cache, "_multi_set", autospec=True) as m:
            await base_cache.multi_set([[Keys.KEY, "value"], [Keys.KEY_1, "value1"]])

            m.assert_called_once_with(
                [(Keys.KEY, "value"), (Keys.KEY_1, "value1")], _conn=None, ttl=None
            )

    async def test_multi_set_ttl_default(self, base_cache):
        base_cache.ttl = 10
        with patch.object(base_cache, "_multi_set", autospec=True) as m:
            await base_cache.multi_set([[Keys.KEY, "value"], [Keys.KEY_1, "value1"]])

            m.assert_called_once_with(
                [(Keys.KEY, "value"), (Keys.KEY_1, "value1")], _conn=None, ttl=10
            )

    async def test_multi_set_ttl_overriden(self, base_cache):
        base_cache.ttl = 10
        with patch.object(base_cache, "_multi_set", autospec=True) as m:
            await base_cache.multi_set([[Keys.KEY, "value"], [Keys.KEY_1, "value1"]], ttl=20)

            m.assert_called_once_with(
                [(Keys.KEY, "value"), (Keys.KEY_1, "value1")], _conn=None, ttl=20
            )

    async def test_multi_set_ttl_none(self, base_cache):
        base_cache.ttl = 10
        with patch.object(base_cache, "_multi_set", autospec=True) as m:
            await base_cache.multi_set([[Keys.KEY, "value"], [Keys.KEY_1, "value1"]], ttl=None)

            m.assert_called_once_with(
                [(Keys.KEY, "value"), (Keys.KEY_1, "value1")], _conn=None, ttl=None
            )


class TestCache:
    """
    This class ensures that all backends behave the same way at logic level. It tries to ensure
    the calls to the necessary methods like serialization and strategies are performed when needed.
    To add a new backend just create the fixture for the new backend and add id as a param for the
    cache fixture.

    The calls to the client are mocked so it doesn't interact with any storage.
    """

    async def asleep(self, *args, **kwargs):
        await asyncio.sleep(0.005)

    async def test_get(self, mock_base_cache):
        await mock_base_cache.get(Keys.KEY)

        mock_base_cache._get.assert_called_with(
            mock_base_cache._build_key(Keys.KEY), encoding=ANY, _conn=ANY
        )
        assert mock_base_cache.plugins[0].pre_get.call_count == 1
        assert mock_base_cache.plugins[0].post_get.call_count == 1

    async def test_get_timeouts(self, mock_base_cache):
        mock_base_cache._get = self.asleep

        with pytest.raises(asyncio.TimeoutError):
            await mock_base_cache.get(Keys.KEY)

    async def test_get_default(self, mock_base_cache):
        mock_base_cache._serializer.loads.return_value = None

        assert await mock_base_cache.get(Keys.KEY, default=1) == 1

    async def test_get_negative_default(self, mock_base_cache):
        mock_base_cache._serializer.loads.return_value = False

        assert await mock_base_cache.get(Keys.KEY) is False

    async def test_set(self, mock_base_cache):
        await mock_base_cache.set(Keys.KEY, "value", ttl=2)

        mock_base_cache._set.assert_called_with(
            mock_base_cache._build_key(Keys.KEY), ANY, ttl=2, _cas_token=None, _conn=ANY
        )
        assert mock_base_cache.plugins[0].pre_set.call_count == 1
        assert mock_base_cache.plugins[0].post_set.call_count == 1

    async def test_set_timeouts(self, mock_base_cache):
        mock_base_cache._set = self.asleep

        with pytest.raises(asyncio.TimeoutError):
            await mock_base_cache.set(Keys.KEY, "value")

    async def test_add(self, mock_base_cache):
        mock_base_cache._exists = AsyncMock(return_value=False)
        await mock_base_cache.add(Keys.KEY, "value", ttl=2)

        key = mock_base_cache._build_key(Keys.KEY)
        mock_base_cache._add.assert_called_with(key, ANY, ttl=2, _conn=ANY)
        assert mock_base_cache.plugins[0].pre_add.call_count == 1
        assert mock_base_cache.plugins[0].post_add.call_count == 1

    async def test_add_timeouts(self, mock_base_cache):
        mock_base_cache._add = self.asleep

        with pytest.raises(asyncio.TimeoutError):
            await mock_base_cache.add(Keys.KEY, "value")

    async def test_mget(self, mock_base_cache):
        await mock_base_cache.multi_get([Keys.KEY, Keys.KEY_1])

        mock_base_cache._multi_get.assert_called_with(
            [mock_base_cache._build_key(Keys.KEY), mock_base_cache._build_key(Keys.KEY_1)],
            encoding=ANY,
            _conn=ANY,
        )
        assert mock_base_cache.plugins[0].pre_multi_get.call_count == 1
        assert mock_base_cache.plugins[0].post_multi_get.call_count == 1

    async def test_mget_timeouts(self, mock_base_cache):
        mock_base_cache._multi_get = self.asleep

        with pytest.raises(asyncio.TimeoutError):
            await mock_base_cache.multi_get(Keys.KEY, "value")

    async def test_mset(self, mock_base_cache):
        await mock_base_cache.multi_set([[Keys.KEY, "value"], [Keys.KEY_1, "value1"]], ttl=2)

        key = mock_base_cache._build_key(Keys.KEY)
        key1 = mock_base_cache._build_key(Keys.KEY_1)
        mock_base_cache._multi_set.assert_called_with(
            [(key, ANY), (key1, ANY)], ttl=2, _conn=ANY)
        assert mock_base_cache.plugins[0].pre_multi_set.call_count == 1
        assert mock_base_cache.plugins[0].post_multi_set.call_count == 1

    async def test_mset_timeouts(self, mock_base_cache):
        mock_base_cache._multi_set = self.asleep

        with pytest.raises(asyncio.TimeoutError):
            await mock_base_cache.multi_set([[Keys.KEY, "value"], [Keys.KEY_1, "value1"]])

    async def test_exists(self, mock_base_cache):
        await mock_base_cache.exists(Keys.KEY)

        mock_base_cache._exists.assert_called_with(mock_base_cache._build_key(Keys.KEY), _conn=ANY)
        assert mock_base_cache.plugins[0].pre_exists.call_count == 1
        assert mock_base_cache.plugins[0].post_exists.call_count == 1

    async def test_exists_timeouts(self, mock_base_cache):
        mock_base_cache._exists = self.asleep

        with pytest.raises(asyncio.TimeoutError):
            await mock_base_cache.exists(Keys.KEY)

    async def test_increment(self, mock_base_cache):
        await mock_base_cache.increment(Keys.KEY, 2)

        key = mock_base_cache._build_key(Keys.KEY)
        mock_base_cache._increment.assert_called_with(key, 2, _conn=ANY)
        assert mock_base_cache.plugins[0].pre_increment.call_count == 1
        assert mock_base_cache.plugins[0].post_increment.call_count == 1

    async def test_increment_timeouts(self, mock_base_cache):
        mock_base_cache._increment = self.asleep

        with pytest.raises(asyncio.TimeoutError):
            await mock_base_cache.increment(Keys.KEY)

    async def test_delete(self, mock_base_cache):
        await mock_base_cache.delete(Keys.KEY)

        mock_base_cache._delete.assert_called_with(mock_base_cache._build_key(Keys.KEY), _conn=ANY)
        assert mock_base_cache.plugins[0].pre_delete.call_count == 1
        assert mock_base_cache.plugins[0].post_delete.call_count == 1

    async def test_delete_timeouts(self, mock_base_cache):
        mock_base_cache._delete = self.asleep

        with pytest.raises(asyncio.TimeoutError):
            await mock_base_cache.delete(Keys.KEY)

    async def test_expire(self, mock_base_cache):
        await mock_base_cache.expire(Keys.KEY, 1)
        key = mock_base_cache._build_key(Keys.KEY)
        mock_base_cache._expire.assert_called_with(key, 1, _conn=ANY)
        assert mock_base_cache.plugins[0].pre_expire.call_count == 1
        assert mock_base_cache.plugins[0].post_expire.call_count == 1

    async def test_expire_timeouts(self, mock_base_cache):
        mock_base_cache._expire = self.asleep

        with pytest.raises(asyncio.TimeoutError):
            await mock_base_cache.expire(Keys.KEY, 0)

    async def test_clear(self, mock_base_cache):
        await mock_base_cache.clear(Keys.KEY)
        mock_base_cache._clear.assert_called_with(mock_base_cache._build_key(Keys.KEY), _conn=ANY)
        assert mock_base_cache.plugins[0].pre_clear.call_count == 1
        assert mock_base_cache.plugins[0].post_clear.call_count == 1

    async def test_clear_timeouts(self, mock_base_cache):
        mock_base_cache._clear = self.asleep

        with pytest.raises(asyncio.TimeoutError):
            await mock_base_cache.clear(Keys.KEY)

    async def test_raw(self, mock_base_cache):
        await mock_base_cache.raw("get", Keys.KEY)
        mock_base_cache._raw.assert_called_with(
            "get", mock_base_cache._build_key(Keys.KEY), encoding=ANY, _conn=ANY
        )
        assert mock_base_cache.plugins[0].pre_raw.call_count == 1
        assert mock_base_cache.plugins[0].post_raw.call_count == 1

    async def test_raw_timeouts(self, mock_base_cache):
        mock_base_cache._raw = self.asleep

        with pytest.raises(asyncio.TimeoutError):
            await mock_base_cache.raw("clear")

    async def test_close(self, mock_base_cache):
        await mock_base_cache.close()
        assert mock_base_cache._close.call_count == 1

    async def test_get_connection(self, mock_base_cache):
        async with mock_base_cache.get_connection():
            pass
        assert mock_base_cache.acquire_conn.call_count == 1
        assert mock_base_cache.release_conn.call_count == 1


@pytest.fixture
def conn(mock_base_cache):
    yield _Conn(mock_base_cache)


class TestConn:
    def test_conn(self, conn, mock_base_cache):
        assert conn._cache == mock_base_cache

    def test_conn_getattr(self, conn, mock_base_cache):
        assert conn.timeout == mock_base_cache.timeout
        assert conn.namespace == mock_base_cache.namespace
        assert conn.serializer is mock_base_cache.serializer

    async def test_conn_context_manager(self, conn):
        async with conn:
            assert conn._cache.acquire_conn.call_count == 1
        conn._cache.release_conn.assert_called_with(conn._cache.acquire_conn.return_value)

    async def test_inject_conn(self, conn):
        conn._conn = "connection"
        conn._cache.dummy = AsyncMock(spec_set=())
        await _Conn._inject_conn("dummy")(conn, "a", b="b")
        conn._cache.dummy.assert_called_with("a", _conn=conn._conn, b="b")



import asyncio

from aiocache import Cache


cache = Cache(Cache.REDIS, endpoint="127.0.0.1", port=6379, namespace="main")


async def redis():
    await cache.set("key", "value")
    await cache.set("expire_me", "value", ttl=10)

    assert await cache.get("key") == "value"
    assert await cache.get("expire_me") == "value"
    assert await cache.raw("ttl", "main:expire_me") > 0


async def test_redis():
    await redis()
    await cache.delete("key")
    await cache.delete("expire_me")
    await cache.close()


if __name__ == "__main__":
    asyncio.run(test_redis())



import pytest

from aiocache.plugins import HitMissRatioPlugin, TimingPlugin


class TestHitMissRatioPlugin:
    @pytest.mark.parametrize(
        "data, ratio",
        [
            ({"testa": 1, "testb": 2, "testc": 3}, 0.6),
            ({"testa": 1, "testz": 0}, 0.2),
            ({}, 0),
            ({"testa": 1, "testb": 2, "testc": 3, "testd": 4, "teste": 5}, 1),
        ],
    )
    async def test_get_hit_miss_ratio(self, memory_cache, data, ratio):
        keys = ["a", "b", "c", "d", "e", "f"]
        memory_cache.plugins = [HitMissRatioPlugin()]
        memory_cache._cache = data

        for key in keys:
            await memory_cache.get(key)

        hits = [x for x in keys if "test" + x in data]
        assert memory_cache.hit_miss_ratio["hits"] == len(hits)
        assert (
            memory_cache.hit_miss_ratio["hit_ratio"]
            == len(hits) / memory_cache.hit_miss_ratio["total"]
        )

    @pytest.mark.parametrize(
        "data, ratio",
        [
            ({"testa": 1, "testb": 2, "testc": 3}, 0.6),
            ({"testa": 1, "testz": 0}, 0.2),
            ({}, 0),
            ({"testa": 1, "testb": 2, "testc": 3, "testd": 4, "teste": 5}, 1),
        ],
    )
    async def test_multi_get_hit_miss_ratio(self, memory_cache, data, ratio):
        keys = ["a", "b", "c", "d", "e", "f"]
        memory_cache.plugins = [HitMissRatioPlugin()]
        memory_cache._cache = data

        for key in keys:
            await memory_cache.multi_get([key])

        hits = [x for x in keys if "test" + x in data]
        assert memory_cache.hit_miss_ratio["hits"] == len(hits)
        assert (
            memory_cache.hit_miss_ratio["hit_ratio"]
            == len(hits) / memory_cache.hit_miss_ratio["total"]
        )

    async def test_set_and_get_using_namespace(self, memory_cache):
        memory_cache.plugins = [HitMissRatioPlugin()]
        key = "A"
        namespace = "test"
        value = 1
        await memory_cache.set(key, value, namespace=namespace)
        result = await memory_cache.get(key, namespace=namespace)
        assert result == value


class TestTimingPlugin:
    @pytest.mark.parametrize(
        "data, ratio",
        [
            ({"testa": 1, "testb": 2, "testc": 3}, 0.6),
            ({"testa": 1, "testz": 0}, 0.2),
            ({}, 0),
            ({"testa": 1, "testb": 2, "testc": 3, "testd": 4, "teste": 5}, 1),
        ],
    )
    async def test_get_avg_min_max(self, memory_cache, data, ratio):
        keys = ["a", "b", "c", "d", "e", "f"]
        memory_cache.plugins = [TimingPlugin()]
        memory_cache._cache = data

        for key in keys:
            await memory_cache.get(key)

        assert "get_max" in memory_cache.profiling
        assert "get_min" in memory_cache.profiling
        assert "get_total" in memory_cache.profiling
        assert "get_avg" in memory_cache.profiling



import asyncio
import json

from marshmallow import Schema, fields, post_load
from aiocache import Cache


class MyType:
    def __init__(self, x, y):
        self.x = x
        self.y = y


class MyTypeSchema(Schema):
    x = fields.Number()
    y = fields.Number()

    @post_load
    def build_object(self, data, **kwargs):
        return MyType(data['x'], data['y'])


def dumps(value):
    return MyTypeSchema().dumps(value)


def loads(value):
    return MyTypeSchema().loads(value)


cache = Cache(Cache.REDIS, namespace="main")


async def serializer_function():
    await cache.set("key", MyType(1, 2), dumps_fn=dumps)

    obj = await cache.get("key", loads_fn=loads)

    assert obj.x == 1
    assert obj.y == 2
    assert await cache.get("key") == json.loads(('{"y": 2.0, "x": 1.0}'))
    assert json.loads(await cache.raw("get", "main:key")) == {"y": 2.0, "x": 1.0}


async def test_serializer_function():
    await serializer_function()
    await cache.delete("key")
    await cache.close()


if __name__ == "__main__":
    asyncio.run(test_serializer_function())



import asyncio

from aiocache import multi_cached, Cache

DICT = {
    'a': "Z",
    'b': "Y",
    'c': "X",
    'd': "W"
}


@multi_cached("ids", cache=Cache.REDIS, namespace="main")
async def multi_cached_ids(ids=None):
    return {id_: DICT[id_] for id_ in ids}


@multi_cached("keys", cache=Cache.REDIS, namespace="main")
async def multi_cached_keys(keys=None):
    return {id_: DICT[id_] for id_ in keys}


cache = Cache(Cache.REDIS, endpoint="127.0.0.1", port=6379, namespace="main")


async def test_multi_cached():
    await multi_cached_ids(ids=("a", "b"))
    await multi_cached_ids(ids=("a", "c"))
    await multi_cached_keys(keys=("d",))

    assert await cache.exists("a")
    assert await cache.exists("b")
    assert await cache.exists("c")
    assert await cache.exists("d")

    await cache.delete("a")
    await cache.delete("b")
    await cache.delete("c")
    await cache.delete("d")
    await cache.close()


if __name__ == "__main__":
    asyncio.run(test_multi_cached())



import asyncio
import functools
import inspect
import logging

from aiocache.base import SENTINEL
from aiocache.factory import Cache, caches
from aiocache.lock import RedLock


logger = logging.getLogger(__name__)


class cached:
    """
    Caches the functions return value into a key generated with module_name, function_name
    and args. The cache is available in the function object as ``<function_name>.cache``.

    In some cases you will need to send more args to configure the cache object.
    An example would be endpoint and port for the Redis cache. You can send those args as
    kwargs and they will be propagated accordingly.

    Only one cache instance is created per decorated call. If you expect high concurrency of
    calls to the same function, you should adapt the pool size as needed.

    Extra args that are injected in the function that you can use to control the cache
    behavior are:

        - ``cache_read``: Controls whether the function call will try to read from cache first or
                          not. Enabled by default.
        - ``cache_write``: Controls whether the function call will try to write in the cache once
                           the result has been retrieved. Enabled by default.
        - ``aiocache_wait_for_write``: Controls whether the call of the function will wait for the
                                       value in the cache to be written. If set to False, the write
                                       happens in the background. Enabled by default

    :param ttl: int seconds to store the function call. Default is None which means no expiration.
    :param key: str value to set as key for the function return. Takes precedence over
        key_builder param. If key and key_builder are not passed, it will use module_name
        + function_name + args + kwargs
    :param namespace: string to use as default prefix for the key used in all operations of
        the backend. Default is None
    :param key_builder: Callable that allows to build the function dynamically. It receives
        the function plus same args and kwargs passed to the function.
        This behavior is necessarily different than ``BaseCache.build_key()``
    :param skip_cache_func: Callable that receives the result after calling the
        wrapped function and should return `True` if the value should skip the
        cache (or `False` to store in the cache).
        e.g. to avoid caching `None` results: `lambda r: r is None`
    :param cache: cache class to use when calling the ``set``/``get`` operations.
        Default is :class:`aiocache.SimpleMemoryCache`.
    :param serializer: serializer instance to use when calling the ``dumps``/``loads``.
        If its None, default one from the cache backend is used.
    :param plugins: list plugins to use when calling the cmd hooks
        Default is pulled from the cache class being used.
    :param alias: str specifying the alias to load the config from. If alias is passed, other
        config parameters are ignored. Same cache identified by alias is used on every call. If
        you need a per function cache, specify the parameters explicitly without using alias.
    :param noself: bool if you are decorating a class function, by default self is also used to
        generate the key. This will result in same function calls done by different class instances
        to use different cache keys. Use noself=True if you want to ignore it.
    """

    def __init__(
        self,
        ttl=SENTINEL,
        key=None,
        namespace=None,
        key_builder=None,
        skip_cache_func=lambda x: False,
        cache=Cache.MEMORY,
        serializer=None,
        plugins=None,
        alias=None,
        noself=False,
        **kwargs,
    ):
        self.ttl = ttl
        self.key = key
        self.key_builder = key_builder
        self.skip_cache_func = skip_cache_func
        self.noself = noself
        self.alias = alias
        self.cache = None

        self._cache = cache
        self._serializer = serializer
        self._namespace = namespace
        self._plugins = plugins
        self._kwargs = kwargs

    def __call__(self, f):
        if self.alias:
            self.cache = caches.get(self.alias)
            for arg in ("serializer", "namespace", "plugins"):
                if getattr(self, f'_{arg}', None) is not None:
                    logger.warning(f"Using cache alias; ignoring '{arg}' argument.")
        else:
            self.cache = _get_cache(
                cache=self._cache,
                serializer=self._serializer,
                namespace=self._namespace,
                plugins=self._plugins,
                **self._kwargs,
            )

        @functools.wraps(f)
        async def wrapper(*args, **kwargs):
            return await self.decorator(f, *args, **kwargs)

        wrapper.cache = self.cache
        return wrapper

    async def decorator(
        self, f, *args, cache_read=True, cache_write=True, aiocache_wait_for_write=True, **kwargs
    ):
        key = self.get_cache_key(f, args, kwargs)

        if cache_read:
            value = await self.get_from_cache(key)
            if value is not None:
                return value

        result = await f(*args, **kwargs)

        if self.skip_cache_func(result):
            return result

        if cache_write:
            if aiocache_wait_for_write:
                await self.set_in_cache(key, result)
            else:
                # TODO: Use aiojobs to avoid warnings.
                asyncio.create_task(self.set_in_cache(key, result))

        return result

    def get_cache_key(self, f, args, kwargs):
        if self.key:
            return self.key
        if self.key_builder:
            return self.key_builder(f, *args, **kwargs)

        return self._key_from_args(f, args, kwargs)

    def _key_from_args(self, func, args, kwargs):
        ordered_kwargs = sorted(kwargs.items())
        return (
            (func.__module__ or "")
            + func.__name__
            + str(args[1:] if self.noself else args)
            + str(ordered_kwargs)
        )

    async def get_from_cache(self, key: str):
        try:
            return await self.cache.get(key)
        except Exception:
            logger.exception("Couldn't retrieve %s, unexpected error", key)
        return None

    async def set_in_cache(self, key, value):
        try:
            await self.cache.set(key, value, ttl=self.ttl)
        except Exception:
            logger.exception("Couldn't set %s in key %s, unexpected error", value, key)


class cached_stampede(cached):
    """
    Caches the functions return value into a key generated with module_name, function_name and args
    while avoids for cache stampede effects.

    In some cases you will need to send more args to configure the cache object.
    An example would be endpoint and port for the Redis cache. You can send those args as
    kwargs and they will be propagated accordingly.

    Only one cache instance is created per decorated function. If you expect high concurrency
    of calls to the same function, you should adapt the pool size as needed.

    :param lease: int seconds to lock function call to avoid cache stampede effects.
        If 0 or None, no locking happens (default is 2). redis and memory backends support
        float ttls
    :param ttl: int seconds to store the function call. Default is None which means no expiration.
    :param key: str value to set as key for the function return. Takes precedence over
        key_from_attr param. If key and key_from_attr are not passed, it will use module_name
        + function_name + args + kwargs
    :param key_from_attr: str arg or kwarg name from the function to use as a key.
    :param namespace: string to use as default prefix for the key used in all operations of
        the backend. Default is None
    :param key_builder: Callable that allows to build the function dynamically. It receives
        the function plus same args and kwargs passed to the function.
        This behavior is necessarily different than ``BaseCache.build_key()``
    :param skip_cache_func: Callable that receives the result after calling the
        wrapped function and should return `True` if the value should skip the
        cache (or `False` to store in the cache).
        e.g. to avoid caching `None` results: `lambda r: r is None`
    :param cache: cache class to use when calling the ``set``/``get`` operations.
        Default is :class:`aiocache.SimpleMemoryCache`.
    :param serializer: serializer instance to use when calling the ``dumps``/``loads``.
        Default is JsonSerializer.
    :param plugins: list plugins to use when calling the cmd hooks
        Default is pulled from the cache class being used.
    :param alias: str specifying the alias to load the config from. If alias is passed,
        other config parameters are ignored. New cache is created every time.
    :param noself: bool if you are decorating a class function, by default self is also used to
        generate the key. This will result in same function calls done by different class instances
        to use different cache keys. Use noself=True if you want to ignore it.
    """

    def __init__(self, lease=2, **kwargs):
        super().__init__(**kwargs)
        self.lease = lease

    async def decorator(self, f, *args, **kwargs):
        key = self.get_cache_key(f, args, kwargs)

        value = await self.get_from_cache(key)
        if value is not None:
            return value

        async with RedLock(self.cache, key, self.lease):
            value = await self.get_from_cache(key)
            if value is not None:
                return value

            result = await f(*args, **kwargs)

            if self.skip_cache_func(result):
                return result

            await self.set_in_cache(key, result)

        return result


def _get_cache(cache=Cache.MEMORY, serializer=None, plugins=None, **cache_kwargs):
    return Cache(cache, serializer=serializer, plugins=plugins, **cache_kwargs)


def _get_args_dict(func, args, kwargs):
    defaults = {
        arg_name: arg.default
        for arg_name, arg in inspect.signature(func).parameters.items()
        if arg.default is not inspect._empty  # TODO: bug prone..
    }
    args_names = func.__code__.co_varnames[: func.__code__.co_argcount]
    return {**defaults, **dict(zip(args_names, args)), **kwargs}


class multi_cached:
    """
    Only supports functions that return dict-like structures. This decorator caches each key/value
    of the dict-like object returned by the function. The dict keys of the returned data should
    match the set of keys that are passed to the decorated callable in an iterable object.
    The name of that argument is passed to this decorator via the parameter
    ``keys_from_attr``. ``keys_from_attr`` can be the name of a positional or keyword argument.

    If the argument specified by ``keys_from_attr`` is an empty list, the cache will be ignored
    and the function will be called. If only some of the keys in ``keys_from_attr``are cached
    (and ``cache_read`` is True) those values will be fetched from the cache, and only the
    uncached keys will be passed to the callable via the argument specified by ``keys_from_attr``.

    By default, the callable's name and call signature are not incorporated into the cache key,
    so if there is another cached function returning a dict with same keys, those keys will be
    overwritten. To avoid this, use a specific ``namespace`` in each cache decorator or pass a
    ``key_builder``.

    If ``key_builder`` is passed, then the values of ``keys_from_attr`` will be transformed
    before requesting them from the cache. Equivalently, the keys in the dict-like mapping
    returned by the decorated callable will be transformed before storing them in the cache.

    The cache is available in the function object as ``<function_name>.cache``.

    Only one cache instance is created per decorated function. If you expect high concurrency
    of calls to the same function, you should adapt the pool size as needed.

    Extra args that are injected in the function that you can use to control the cache
    behavior are:

        - ``cache_read``: Controls whether the function call will try to read from cache first or
                          not. Enabled by default.
        - ``cache_write``: Controls whether the function call will try to write in the cache once
                           the result has been retrieved. Enabled by default.
        - ``aiocache_wait_for_write``: Controls whether the call of the function will wait for the
                                       value in the cache to be written. If set to False, the write
                                       happens in the background. Enabled by default

    :param keys_from_attr: name of the arg or kwarg in the decorated callable that contains
        an iterable that yields the keys returned by the decorated callable.
    :param namespace: string to use as default prefix for the key used in all operations of
        the backend. Default is None
    :param key_builder: Callable that enables mapping the decorated function's keys to the keys
        used by the cache. Receives a key from the iterable corresponding to
        ``keys_from_attr``, the decorated callable, and the positional and keyword arguments
        that were passed to the decorated callable. This behavior is necessarily different than
        ``BaseCache.build_key()`` and the call signature differs from ``cached.key_builder``.
    :param skip_cache_func: Callable that receives both key and value and returns True
        if that key-value pair should not be cached (or False to store in cache).
        The keys and values to be passed are taken from the wrapped function result.
    :param ttl: int seconds to store the keys. Default is 0 which means no expiration.
    :param cache: cache class to use when calling the ``multi_set``/``multi_get`` operations.
        Default is :class:`aiocache.SimpleMemoryCache`.
    :param serializer: serializer instance to use when calling the ``dumps``/``loads``.
        If its None, default one from the cache backend is used.
    :param plugins: plugins to use when calling the cmd hooks
        Default is pulled from the cache class being used.
    :param alias: str specifying the alias to load the config from. If alias is passed,
        other config parameters are ignored. Same cache identified by alias is used on
        every call. If you need a per function cache, specify the parameters explicitly
        without using alias.
    """

    def __init__(
        self,
        keys_from_attr,
        namespace=None,
        key_builder=None,
        skip_cache_func=lambda k, v: False,
        ttl=SENTINEL,
        cache=Cache.MEMORY,
        serializer=None,
        plugins=None,
        alias=None,
        **kwargs,
    ):
        self.keys_from_attr = keys_from_attr
        self.key_builder = key_builder or (lambda key, f, *args, **kwargs: key)
        self.skip_cache_func = skip_cache_func
        self.ttl = ttl
        self.alias = alias
        self.cache = None

        self._cache = cache
        self._serializer = serializer
        self._namespace = namespace
        self._plugins = plugins
        self._kwargs = kwargs

    def __call__(self, f):
        if self.alias:
            self.cache = caches.get(self.alias)
            for arg in ("serializer", "namespace", "plugins"):
                if getattr(self, f'_{arg}', None) is not None:
                    logger.warning(f"Using cache alias; ignoring '{arg}' argument.")
        else:
            self.cache = _get_cache(
                cache=self._cache,
                serializer=self._serializer,
                namespace=self._namespace,
                plugins=self._plugins,
                **self._kwargs,
            )

        @functools.wraps(f)
        async def wrapper(*args, **kwargs):
            return await self.decorator(f, *args, **kwargs)

        wrapper.cache = self.cache
        return wrapper

    async def decorator(
        self, f, *args, cache_read=True, cache_write=True, aiocache_wait_for_write=True, **kwargs
    ):
        missing_keys = []
        partial = {}
        keys, new_args, args_index = self.get_cache_keys(f, args, kwargs)

        if cache_read:
            values = await self.get_from_cache(*keys)
            for key, value in zip(keys, values):
                if value is None:
                    missing_keys.append(key)
                else:
                    partial[key] = value
            if values and None not in values:
                return partial
        else:
            missing_keys = list(keys)

        if args_index > -1:
            new_args[args_index] = missing_keys
        else:
            kwargs[self.keys_from_attr] = missing_keys

        result = await f(*new_args, **kwargs)
        result.update(partial)

        to_cache = {k: v for k, v in result.items() if not self.skip_cache_func(k, v)}

        if not to_cache:
            return result

        if cache_write:
            if aiocache_wait_for_write:
                await self.set_in_cache(to_cache, f, args, kwargs)
            else:
                # TODO: Use aiojobs to avoid warnings.
                asyncio.create_task(self.set_in_cache(to_cache, f, args, kwargs))

        return result

    def get_cache_keys(self, f, args, kwargs):
        args_dict = _get_args_dict(f, args, kwargs)
        keys = args_dict.get(self.keys_from_attr, []) or []
        keys = [self.key_builder(key, f, *args, **kwargs) for key in keys]

        args_names = f.__code__.co_varnames[: f.__code__.co_argcount]
        new_args = list(args)
        keys_index = -1
        if self.keys_from_attr in args_names and self.keys_from_attr not in kwargs:
            keys_index = args_names.index(self.keys_from_attr)
            new_args[keys_index] = keys

        return keys, new_args, keys_index

    async def get_from_cache(self, *keys):
        if not keys:
            return []
        try:
            values = await self.cache.multi_get(keys)
            return values
        except Exception:
            logger.exception("Couldn't retrieve %s, unexpected error", keys)
            return [None] * len(keys)

    async def set_in_cache(self, result, fn, fn_args, fn_kwargs):
        try:
            await self.cache.multi_set(
                [(self.key_builder(k, fn, *fn_args, **fn_kwargs), v) for k, v in result.items()],
                ttl=self.ttl,
            )
        except Exception:
            logger.exception("Couldn't set %s, unexpected error", result)



import logging

from .serializers import (
    BaseSerializer,
    JsonSerializer,
    NullSerializer,
    PickleSerializer,
    StringSerializer,
)

logger = logging.getLogger(__name__)


try:
    import msgpack
except ImportError:
    logger.debug("msgpack not installed, MsgPackSerializer unavailable")
else:
    from .serializers import MsgPackSerializer

    del msgpack


__all__ = [
    "BaseSerializer",
    "NullSerializer",
    "StringSerializer",
    "PickleSerializer",
    "JsonSerializer",
    "MsgPackSerializer",
]



import asyncio
import logging
from datetime import datetime
from aiohttp import web
from aiocache import cached
from aiocache.serializers import JsonSerializer


@cached(key="function_key", serializer=JsonSerializer())
async def time():
    return {"time": datetime.now().isoformat()}


async def handle(request):
    return web.json_response(await time())


# It is also possible to cache the whole route, but for this you will need to
# override `cached.get_from_cache` and regenerate the response since aiohttp
# forbids reusing responses
class CachedOverride(cached):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    async def get_from_cache(self, key):
        try:
            value = await self.cache.get(key)
            if type(value) == web.Response:
                return web.Response(
                    body=value.body,
                    status=value.status,
                    reason=value.reason,
                    headers=value.headers,
                )
            return value
        except Exception:
            logging.exception("Couldn't retrieve %s, unexpected error", key)
        return None


@CachedOverride(key="route_key", serializer=JsonSerializer())
async def handle2(request):
    return web.json_response(await asyncio.sleep(3))


if __name__ == "__main__":
    app = web.Application()
    app.router.add_get('/handle', handle)
    app.router.add_get('/handle2', handle2)

    web.run_app(app)



import logging
from typing import Dict, Type

from .backends.memory import SimpleMemoryCache
from .base import BaseCache

__version__ = "0.12.2"

logger = logging.getLogger(__name__)

AIOCACHE_CACHES: Dict[str, Type[BaseCache]] = {SimpleMemoryCache.NAME: SimpleMemoryCache}

try:
    import redis
except ImportError:
    logger.debug("redis not installed, RedisCache unavailable")
else:
    from aiocache.backends.redis import RedisCache

    AIOCACHE_CACHES[RedisCache.NAME] = RedisCache
    del redis

try:
    import aiomcache
except ImportError:
    logger.debug("aiomcache not installed, Memcached unavailable")
else:
    from aiocache.backends.memcached import MemcachedCache

    AIOCACHE_CACHES[MemcachedCache.NAME] = MemcachedCache
    del aiomcache

from .decorators import cached, cached_stampede, multi_cached  # noqa: E402,I202
from .factory import Cache, caches  # noqa: E402


__all__ = (
    "caches",
    "Cache",
    "cached",
    "cached_stampede",
    "multi_cached",
    *(c.__name__ for c in AIOCACHE_CACHES.values()),
)



from unittest.mock import create_autospec

import pytest

from aiocache.base import API, BaseCache
from aiocache.plugins import BasePlugin, HitMissRatioPlugin, TimingPlugin
from ..utils import Keys


class TestBasePlugin:
    async def test_interface_methods(self):
        for method in API.CMDS:
            pre = await getattr(BasePlugin, "pre_{}".format(method.__name__))(None)
            assert pre is None
            post = await getattr(BasePlugin, "post_{}".format(method.__name__))(None)
            assert post is None

    async def test_do_nothing(self):
        assert await BasePlugin().do_nothing() is None


class TestTimingPlugin:
    async def test_save_time(self, mock_cache):
        do_save_time = TimingPlugin().save_time("get")
        await do_save_time("self", mock_cache, took=1)
        await do_save_time("self", mock_cache, took=2)

        assert mock_cache.profiling["get_total"] == 2
        assert mock_cache.profiling["get_max"] == 2
        assert mock_cache.profiling["get_min"] == 1
        assert mock_cache.profiling["get_avg"] == 1.5

    async def test_save_time_post_set(self, mock_cache):
        await TimingPlugin().post_set(mock_cache, took=1)
        await TimingPlugin().post_set(mock_cache, took=2)

        assert mock_cache.profiling["set_total"] == 2
        assert mock_cache.profiling["set_max"] == 2
        assert mock_cache.profiling["set_min"] == 1
        assert mock_cache.profiling["set_avg"] == 1.5

    async def test_interface_methods(self):
        for method in API.CMDS:
            assert hasattr(TimingPlugin, "pre_{}".format(method.__name__))
            assert hasattr(TimingPlugin, "post_{}".format(method.__name__))


class TestHitMissRatioPlugin:
    @pytest.fixture
    def plugin(self):
        return HitMissRatioPlugin()

    async def test_post_get(self, plugin):
        client = create_autospec(BaseCache, instance=True)
        await plugin.post_get(client, Keys.KEY)

        assert client.hit_miss_ratio["hits"] == 0
        assert client.hit_miss_ratio["total"] == 1
        assert client.hit_miss_ratio["hit_ratio"] == 0

        await plugin.post_get(client, Keys.KEY, ret="value")
        assert client.hit_miss_ratio["hits"] == 1
        assert client.hit_miss_ratio["total"] == 2
        assert client.hit_miss_ratio["hit_ratio"] == 0.5

    async def test_post_multi_get(self, plugin):
        client = create_autospec(BaseCache, instance=True)
        await plugin.post_multi_get(client, [Keys.KEY, Keys.KEY_1], ret=[None, None])

        assert client.hit_miss_ratio["hits"] == 0
        assert client.hit_miss_ratio["total"] == 2
        assert client.hit_miss_ratio["hit_ratio"] == 0

        await plugin.post_multi_get(client, [Keys.KEY, Keys.KEY_1], ret=["value", "random"])
        assert client.hit_miss_ratio["hits"] == 2
        assert client.hit_miss_ratio["total"] == 4
        assert client.hit_miss_ratio["hit_ratio"] == 0.5



"""
Example of caching using aiocache package:

    /: Does a 3 seconds sleep. Only the first time because its using the `cached` decorator
    /reuse: Returns the data stored in "main" endpoint
"""

import asyncio

from sanic import Sanic
from sanic.response import json
from sanic.log import logger
from aiocache import cached, Cache
from aiocache.serializers import JsonSerializer

app = Sanic(__name__)


@cached(key="my_custom_key", serializer=JsonSerializer())
async def expensive_call():
    logger.info("Expensive has been called")
    await asyncio.sleep(3)
    return {"test": True}


async def reuse_data():
    cache = Cache(serializer=JsonSerializer())  # Not ideal to define here
    data = await cache.get("my_custom_key")  # Note the key is defined in `cached` decorator
    return data


@app.route("/")
async def main(request):
    logger.info("Received GET /")
    return json(await expensive_call())


@app.route("/reuse")
async def reuse(request):
    logger.info("Received GET /reuse")
    return json(await reuse_data())


app.run(host="0.0.0.0", port=8000)



import platform
import time

import aiomcache
import pytest
import redis.asyncio as redis


@pytest.fixture
async def redis_client() -> redis.Redis:
    async with redis.Redis(host="127.0.0.1", port=6379, max_connections=1) as r:
        yield r


@pytest.mark.skipif(platform.python_implementation() == "PyPy", reason="Too slow")
class TestRedis:
    async def test_redis_getsetdel(self, redis_client, redis_cache):
        N = 10000
        redis_total_time = 0
        for _n in range(N):
            start = time.time()
            await redis_client.set("hi", "value")
            await redis_client.get("hi")
            await redis_client.delete("hi")
            redis_total_time += time.time() - start

        aiocache_total_time = 0
        for _n in range(N):
            start = time.time()
            await redis_cache.set("hi", "value", timeout=0)
            await redis_cache.get("hi", timeout=0)
            await redis_cache.delete("hi", timeout=0)
            aiocache_total_time += time.time() - start

        print(
            "\n{:0.2f}/{:0.2f}: {:0.2f}".format(
                aiocache_total_time, redis_total_time, aiocache_total_time / redis_total_time
            )
        )
        print("aiocache avg call: {:0.5f}s".format(aiocache_total_time / N))
        print("redis    avg call: {:0.5f}s".format(redis_total_time / N))
        assert aiocache_total_time / redis_total_time < 1.35

    async def test_redis_multigetsetdel(self, redis_client, redis_cache):
        N = 5000
        redis_total_time = 0
        values = ["a", "b", "c", "d", "e", "f"]
        for _n in range(N):
            start = time.time()
            await redis_client.mset({x: x for x in values})
            await redis_client.mget(values)
            for k in values:
                await redis_client.delete(k)
            redis_total_time += time.time() - start

        aiocache_total_time = 0
        for _n in range(N):
            start = time.time()
            await redis_cache.multi_set([(x, x) for x in values], timeout=0)
            await redis_cache.multi_get(values, timeout=0)
            for k in values:
                await redis_cache.delete(k, timeout=0)
            aiocache_total_time += time.time() - start

        print(
            "\n{:0.2f}/{:0.2f}: {:0.2f}".format(
                aiocache_total_time, redis_total_time, aiocache_total_time / redis_total_time
            )
        )
        print("aiocache avg call: {:0.5f}s".format(aiocache_total_time / N))
        print("redis_client    avg call: {:0.5f}s".format(redis_total_time / N))
        assert aiocache_total_time / redis_total_time < 1.35


@pytest.fixture
async def aiomcache_pool():
    client = aiomcache.Client("127.0.0.1", 11211, pool_size=1)
    yield client
    await client.close()


@pytest.mark.skipif(platform.python_implementation() == "PyPy", reason="Too slow")
class TestMemcached:
    async def test_memcached_getsetdel(self, aiomcache_pool, memcached_cache):
        N = 10000
        aiomcache_total_time = 0
        for _n in range(N):
            start = time.time()
            await aiomcache_pool.set(b"hi", b"value")
            await aiomcache_pool.get(b"hi")
            await aiomcache_pool.delete(b"hi")
            aiomcache_total_time += time.time() - start

        aiocache_total_time = 0
        for _n in range(N):
            start = time.time()
            await memcached_cache.set("hi", "value", timeout=0)
            await memcached_cache.get("hi", timeout=0)
            await memcached_cache.delete("hi", timeout=0)
            aiocache_total_time += time.time() - start

        print(
            "\n{:0.2f}/{:0.2f}: {:0.2f}".format(
                aiocache_total_time,
                aiomcache_total_time,
                aiocache_total_time / aiomcache_total_time,
            )
        )
        print("aiocache avg call: {:0.5f}s".format(aiocache_total_time / N))
        print("aiomcache avg call: {:0.5f}s".format(aiomcache_total_time / N))
        assert aiocache_total_time / aiomcache_total_time < 1.40

    async def test_memcached_multigetsetdel(self, aiomcache_pool, memcached_cache):
        N = 2000
        aiomcache_total_time = 0
        values = [b"a", b"b", b"c", b"d", b"e", b"f"]
        for _n in range(N):
            start = time.time()
            for k in values:
                await aiomcache_pool.set(k, k)
            await aiomcache_pool.multi_get(*values)
            for k in values:
                await aiomcache_pool.delete(k)
            aiomcache_total_time += time.time() - start

        aiocache_total_time = 0
        values = ["a", "b", "c", "d", "e", "f"]
        for _n in range(N):
            start = time.time()
            await memcached_cache.multi_set([(x, x) for x in values], timeout=0)
            await memcached_cache.multi_get(values, timeout=0)
            for k in values:
                await memcached_cache.delete(k, timeout=0)
            aiocache_total_time += time.time() - start

        print(
            "\n{:0.2f}/{:0.2f}: {:0.2f}".format(
                aiocache_total_time,
                aiomcache_total_time,
                aiocache_total_time / aiomcache_total_time,
            )
        )
        print("aiocache avg call: {:0.5f}s".format(aiocache_total_time / N))
        print("aiomcache avg call: {:0.5f}s".format(aiomcache_total_time / N))
        assert aiocache_total_time / aiomcache_total_time < 1.40



import pytest

from aiocache import Cache


@pytest.fixture
async def redis_cache():
    # redis connection pool raises ConnectionError but doesn't wait for conn reuse
    #  when exceeding max pool size.
    async with Cache(Cache.REDIS, namespace="test", pool_max_size=1) as cache:
        yield cache


@pytest.fixture
async def memcached_cache():
    async with Cache(Cache.MEMCACHED, namespace="test", pool_size=1) as cache:
        yield cache



import asyncio

from aiocache import caches, Cache
from aiocache.serializers import StringSerializer, PickleSerializer

caches.set_config({
    'default': {
        'cache': "aiocache.SimpleMemoryCache",
        'serializer': {
            'class': "aiocache.serializers.StringSerializer"
        }
    },
    'redis_alt': {
        'cache': "aiocache.RedisCache",
        'endpoint': "127.0.0.1",
        'port': 6379,
        'timeout': 1,
        'serializer': {
            'class': "aiocache.serializers.PickleSerializer"
        },
        'plugins': [
            {'class': "aiocache.plugins.HitMissRatioPlugin"},
            {'class': "aiocache.plugins.TimingPlugin"}
        ]
    }
})


async def default_cache():
    cache = caches.get('default')   # This always returns the same instance
    await cache.set("key", "value")

    assert await cache.get("key") == "value"
    assert isinstance(cache, Cache.MEMORY)
    assert isinstance(cache.serializer, StringSerializer)


async def alt_cache():
    # This generates a new instance every time! You can also use
    # `caches.create("alt", namespace="test", etc...)` to override extra args
    cache = caches.create("redis_alt")
    await cache.set("key", "value")

    assert await cache.get("key") == "value"
    assert isinstance(cache, Cache.REDIS)
    assert isinstance(cache.serializer, PickleSerializer)
    assert len(cache.plugins) == 2
    assert cache.endpoint == "127.0.0.1"
    assert cache.timeout == 1
    assert cache.port == 6379
    await cache.close()


async def test_alias():
    await default_cache()
    await alt_cache()

    cache = Cache(Cache.REDIS)
    await cache.delete("key")
    await cache.close()

    await caches.get("default").close()


if __name__ == "__main__":
    asyncio.run(test_alias())



import asyncio
import logging
import random

from aiocache import Cache
from aiocache.lock import OptimisticLock, OptimisticLockError


logger = logging.getLogger(__name__)
cache = Cache(Cache.REDIS, endpoint='127.0.0.1', port=6379, namespace='main')


async def expensive_function():
    logger.warning('Expensive is being executed...')
    await asyncio.sleep(random.uniform(0, 2))
    return 'result'


async def my_view():

    async with OptimisticLock(cache, 'key') as lock:
        result = await expensive_function()
        try:
            await lock.cas(result)
        except OptimisticLockError:
            logger.warning(
                'I failed setting the value because it is different since the lock started!')
        return result


async def concurrent():
    await cache.set('key', 'initial_value')
    # All three calls will read 'initial_value' as the value to check and only
    # the first one finishing will succeed because the others, when trying to set
    # the value, will see that the value is not the same as when the lock started
    await asyncio.gather(my_view(), my_view(), my_view())


async def test_redis():
    await concurrent()
    await cache.delete("key")
    await cache.close()


if __name__ == '__main__':
    asyncio.run(test_redis())



from unittest.mock import ANY, AsyncMock, create_autospec, patch

import pytest
from redis.asyncio.client import Pipeline
from redis.exceptions import ResponseError

from aiocache.backends.redis import RedisBackend, RedisCache
from aiocache.base import BaseCache, _ensure_key
from aiocache.serializers import JsonSerializer
from ...utils import Keys


@pytest.fixture
def redis():
    redis = RedisBackend()
    with patch.object(redis, "client", autospec=True) as m:
        # These methods actually return an awaitable.
        for method in (
            "eval", "expire", "get", "psetex", "setex", "execute_command", "exists",
            "incrby", "persist", "delete", "keys", "flushdb",
        ):
            setattr(m, method, AsyncMock(return_value=None, spec_set=()))
        m.mget = AsyncMock(return_value=[None], spec_set=())
        m.set = AsyncMock(return_value=True, spec_set=())

        m.pipeline.return_value = create_autospec(Pipeline, instance=True)
        m.pipeline.return_value.__aenter__.return_value = m.pipeline.return_value
        yield redis


class TestRedisBackend:
    default_redis_kwargs = {
        "host": "127.0.0.1",
        "port": 6379,
        "db": 0,
        "password": None,
        "socket_connect_timeout": None,
        "decode_responses": False,
        "max_connections": None,
    }

    @patch("redis.asyncio.Redis", name="mock_class", autospec=True)
    def test_setup(self, mock_class):
        redis_backend = RedisBackend()
        kwargs = self.default_redis_kwargs.copy()
        mock_class.assert_called_with(**kwargs)
        assert redis_backend.endpoint == "127.0.0.1"
        assert redis_backend.port == 6379
        assert redis_backend.db == 0
        assert redis_backend.password is None
        assert redis_backend.pool_max_size is None

    @patch("redis.asyncio.Redis", name="mock_class", autospec=True)
    def test_setup_override(self, mock_class):
        override = {"db": 2, "password": "pass"}
        redis_backend = RedisBackend(**override)

        kwargs = self.default_redis_kwargs.copy()
        kwargs.update(override)
        mock_class.assert_called_with(**kwargs)

        assert redis_backend.endpoint == "127.0.0.1"
        assert redis_backend.port == 6379
        assert redis_backend.db == 2
        assert redis_backend.password == "pass"

    @patch("redis.asyncio.Redis", name="mock_class", autospec=True)
    def test_setup_casts(self, mock_class):
        override = {
            "db": "2",
            "port": "6379",
            "pool_max_size": "10",
            "create_connection_timeout": "1.5",
        }
        redis_backend = RedisBackend(**override)

        kwargs = self.default_redis_kwargs.copy()
        kwargs.update({
            "db": 2,
            "port": 6379,
            "max_connections": 10,
            "socket_connect_timeout": 1.5,
        })
        mock_class.assert_called_with(**kwargs)

        assert redis_backend.db == 2
        assert redis_backend.port == 6379
        assert redis_backend.pool_max_size == 10
        assert redis_backend.create_connection_timeout == 1.5

    async def test_get(self, redis):
        redis.client.get.return_value = b"value"
        assert await redis._get(Keys.KEY) == "value"
        redis.client.get.assert_called_with(Keys.KEY)

    async def test_gets(self, mocker, redis):
        mocker.spy(redis, "_get")
        await redis._gets(Keys.KEY)
        redis._get.assert_called_with(Keys.KEY, encoding="utf-8", _conn=ANY)

    async def test_set(self, redis):
        await redis._set(Keys.KEY, "value")
        redis.client.set.assert_called_with(Keys.KEY, "value")

        await redis._set(Keys.KEY, "value", ttl=1)
        redis.client.setex.assert_called_with(Keys.KEY, 1, "value")

    async def test_set_cas_token(self, mocker, redis):
        mocker.spy(redis, "_cas")
        await redis._set(Keys.KEY, "value", _cas_token="old_value", _conn=redis.client)
        redis._cas.assert_called_with(
            Keys.KEY, "value", "old_value", ttl=None, _conn=redis.client
        )

    async def test_cas(self, mocker, redis):
        mocker.spy(redis, "_raw")
        await redis._cas(Keys.KEY, "value", "old_value", ttl=10, _conn=redis.client)
        redis._raw.assert_called_with(
            "eval",
            redis.CAS_SCRIPT,
            1,
            *[Keys.KEY, "value", "old_value", "EX", 10],
            _conn=redis.client,
        )

    async def test_cas_float_ttl(self, mocker, redis):
        mocker.spy(redis, "_raw")
        await redis._cas(Keys.KEY, "value", "old_value", ttl=0.1, _conn=redis.client)
        redis._raw.assert_called_with(
            "eval",
            redis.CAS_SCRIPT,
            1,
            *[Keys.KEY, "value", "old_value", "PX", 100],
            _conn=redis.client,
        )

    async def test_multi_get(self, redis):
        await redis._multi_get([Keys.KEY, Keys.KEY_1])
        redis.client.mget.assert_called_with(Keys.KEY, Keys.KEY_1)

    async def test_multi_set(self, redis):
        await redis._multi_set([(Keys.KEY, "value"), (Keys.KEY_1, "random")])
        redis.client.execute_command.assert_called_with(
            "MSET", Keys.KEY, "value", Keys.KEY_1, "random"
        )

    async def test_multi_set_with_ttl(self, redis):
        await redis._multi_set([(Keys.KEY, "value"), (Keys.KEY_1, "random")], ttl=1)
        assert redis.client.pipeline.call_count == 1
        pipeline = redis.client.pipeline.return_value
        pipeline.execute_command.assert_called_with(
            "MSET", Keys.KEY, "value", Keys.KEY_1, "random"
        )
        pipeline.expire.assert_any_call(Keys.KEY, time=1)
        pipeline.expire.assert_any_call(Keys.KEY_1, time=1)
        assert pipeline.execute.call_count == 1

    async def test_add(self, redis):
        await redis._add(Keys.KEY, "value")
        redis.client.set.assert_called_with(Keys.KEY, "value", nx=True, ex=None)

        await redis._add(Keys.KEY, "value", 1)
        redis.client.set.assert_called_with(Keys.KEY, "value", nx=True, ex=1)

    async def test_add_existing(self, redis):
        redis.client.set.return_value = False
        with pytest.raises(ValueError):
            await redis._add(Keys.KEY, "value")

    async def test_add_float_ttl(self, redis):
        await redis._add(Keys.KEY, "value", 0.1)
        redis.client.set.assert_called_with(Keys.KEY, "value", nx=True, px=100)

    async def test_exists(self, redis):
        redis.client.exists.return_value = 1
        await redis._exists(Keys.KEY)
        redis.client.exists.assert_called_with(Keys.KEY)

    async def test_increment(self, redis):
        await redis._increment(Keys.KEY, delta=2)
        redis.client.incrby.assert_called_with(Keys.KEY, 2)

    async def test_increment_typerror(self, redis):
        redis.client.incrby.side_effect = ResponseError("msg")
        with pytest.raises(TypeError):
            await redis._increment(Keys.KEY, delta=2)
        redis.client.incrby.assert_called_with(Keys.KEY, 2)

    async def test_expire(self, redis):
        await redis._expire(Keys.KEY, 1)
        redis.client.expire.assert_called_with(Keys.KEY, 1)
        await redis._increment(Keys.KEY, 2)

    async def test_expire_0_ttl(self, redis):
        await redis._expire(Keys.KEY, ttl=0)
        redis.client.persist.assert_called_with(Keys.KEY)

    async def test_delete(self, redis):
        await redis._delete(Keys.KEY)
        redis.client.delete.assert_called_with(Keys.KEY)

    async def test_clear(self, redis):
        redis.client.keys.return_value = ["nm:a", "nm:b"]
        await redis._clear("nm")
        redis.client.delete.assert_called_with("nm:a", "nm:b")

    async def test_clear_no_keys(self, redis):
        redis.client.keys.return_value = []
        await redis._clear("nm")
        redis.client.delete.assert_not_called()

    async def test_clear_no_namespace(self, redis):
        await redis._clear()
        assert redis.client.flushdb.call_count == 1

    async def test_raw(self, redis):
        await redis._raw("get", Keys.KEY)
        await redis._raw("set", Keys.KEY, 1)
        redis.client.get.assert_called_with(Keys.KEY)
        redis.client.set.assert_called_with(Keys.KEY, 1)

    async def test_redlock_release(self, mocker, redis):
        mocker.spy(redis, "_raw")
        await redis._redlock_release(Keys.KEY, "random")
        redis._raw.assert_called_with("eval", redis.RELEASE_SCRIPT, 1, Keys.KEY, "random")

    async def test_close(self, redis):
        await redis._close()
        assert redis.client.close.call_count == 1


class TestRedisCache:
    @pytest.fixture
    def set_test_namespace(self, redis_cache):
        redis_cache.namespace = "test"
        yield
        redis_cache.namespace = None

    def test_name(self):
        assert RedisCache.NAME == "redis"

    def test_inheritance(self):
        assert isinstance(RedisCache(), BaseCache)

    def test_default_serializer(self):
        assert isinstance(RedisCache().serializer, JsonSerializer)

    @pytest.mark.parametrize(
        "path,expected", [("", {}), ("/", {}), ("/1", {"db": "1"}), ("/1/2/3", {"db": "1"})]
    )
    def test_parse_uri_path(self, path, expected):
        assert RedisCache().parse_uri_path(path) == expected

    @pytest.mark.parametrize(
        "namespace, expected",
        ([None, "test:" + _ensure_key(Keys.KEY)], ["", _ensure_key(Keys.KEY)], ["my_ns", "my_ns:" + _ensure_key(Keys.KEY)]),  # noqa: B950
    )
    def test_build_key_double_dot(self, set_test_namespace, redis_cache, namespace, expected):
        assert redis_cache.build_key(Keys.KEY, namespace=namespace) == expected

    def test_build_key_no_namespace(self, redis_cache):
        assert redis_cache.build_key(Keys.KEY, namespace=None) == Keys.KEY



import pytest

from aiocache import Cache
from aiocache.backends.memory import SimpleMemoryCache


class TestCache:
    async def test_from_url_memory(self):
        async with Cache.from_url("memory://") as cache:
            assert isinstance(cache, SimpleMemoryCache)

    def test_from_url_memory_no_endpoint(self):
        with pytest.raises(TypeError):
            Cache.from_url("memory://endpoint:10")

    @pytest.mark.redis
    async def test_from_url_redis(self):
        from aiocache.backends.redis import RedisCache

        url = ("redis://endpoint:1000/0/?password=pass"
               + "&pool_max_size=50&create_connection_timeout=20")

        async with Cache.from_url(url) as cache:
            assert isinstance(cache, RedisCache)
            assert cache.endpoint == "endpoint"
            assert cache.port == 1000
            assert cache.password == "pass"
            assert cache.pool_max_size == 50
            assert cache.create_connection_timeout == 20

    @pytest.mark.memcached
    async def test_from_url_memcached(self):
        from aiocache.backends.memcached import MemcachedCache

        url = "memcached://endpoint:1000?pool_size=10"

        async with Cache.from_url(url) as cache:
            assert isinstance(cache, MemcachedCache)
            assert cache.endpoint == "endpoint"
            assert cache.port == 1000
            assert cache.pool_size == 10

    @pytest.mark.parametrize(
        "scheme",
        (pytest.param("redis", marks=pytest.mark.redis),
         "memory",
         pytest.param("memcached", marks=pytest.mark.memcached),
         ))
    def test_from_url_unexpected_param(self, scheme):
        with pytest.raises(TypeError):
            Cache.from_url("{}://?arg1=arg1".format(scheme))



import asyncio

from collections import namedtuple
from aiocache import Cache
from aiocache.serializers import PickleSerializer


MyObject = namedtuple("MyObject", ["x", "y"])
cache = Cache(Cache.REDIS, serializer=PickleSerializer(), namespace="main")


async def complex_object():
    obj = MyObject(x=1, y=2)
    await cache.set("key", obj)
    my_object = await cache.get("key")

    assert my_object.x == 1
    assert my_object.y == 2


async def test_python_object():
    await complex_object()
    await cache.delete("key")
    await cache.close()


if __name__ == "__main__":
    asyncio.run(test_python_object())



import logging
import pickle  # noqa: S403
from typing import Any, Optional

logger = logging.getLogger(__name__)

try:
    import ujson as json  # noqa: I900
except ImportError:
    logger.debug("ujson module not found, using json")
    import json  # type: ignore[no-redef]

try:
    import msgpack
except ImportError:
    msgpack = None
    logger.debug("msgpack not installed, MsgPackSerializer unavailable")


_NOT_SET = object()


class BaseSerializer:

    DEFAULT_ENCODING: Optional[str] = "utf-8"

    def __init__(self, *args, encoding=_NOT_SET, **kwargs):
        self.encoding = self.DEFAULT_ENCODING if encoding is _NOT_SET else encoding
        super().__init__(*args, **kwargs)

    # TODO(PY38): Positional-only
    def dumps(self, value: Any) -> str:
        raise NotImplementedError("dumps method must be implemented")

    # TODO(PY38): Positional-only
    def loads(self, value: str) -> Any:
        raise NotImplementedError("loads method must be implemented")


class NullSerializer(BaseSerializer):
    """
    This serializer does nothing. Its only recommended to be used by
    :class:`aiocache.SimpleMemoryCache` because for other backends it will
    produce incompatible data unless you work only with str types because it
    store data as is.

    DISCLAIMER: Be careful with mutable types and memory storage. The following
    behavior is considered normal (same as ``functools.lru_cache``)::

        cache = Cache()
        my_list = [1]
        await cache.set("key", my_list)
        my_list.append(2)
        await cache.get("key")  # Will return [1, 2]
    """

    def dumps(self, value):
        """
        Returns the same value
        """
        return value

    def loads(self, value):
        """
        Returns the same value
        """
        return value


class StringSerializer(BaseSerializer):
    """
    Converts all input values to str. All return values are also str. Be
    careful because this means that if you store an ``int(1)``, you will get
    back '1'.

    The transformation is done by just casting to str in the ``dumps`` method.

    If you want to keep python types, use ``PickleSerializer``. ``JsonSerializer``
    may also be useful to keep type of symple python types.
    """

    def dumps(self, value):
        """
        Serialize the received value casting it to str.

        :param value: obj Anything support cast to str
        :returns: str
        """
        return str(value)

    def loads(self, value):
        """
        Returns value back without transformations
        """
        return value


class PickleSerializer(BaseSerializer):
    """
    Transform data to bytes using pickle.dumps and pickle.loads to retrieve it back.
    """

    DEFAULT_ENCODING = None

    def __init__(self, *args, protocol=pickle.DEFAULT_PROTOCOL, **kwargs):
        super().__init__(*args, **kwargs)
        self.protocol = protocol

    def dumps(self, value):
        """
        Serialize the received value using ``pickle.dumps``.

        :param value: obj
        :returns: bytes
        """
        return pickle.dumps(value, protocol=self.protocol)

    def loads(self, value):
        """
        Deserialize value using ``pickle.loads``.

        :param value: bytes
        :returns: obj
        """
        if value is None:
            return None
        return pickle.loads(value)  # noqa: S301


class JsonSerializer(BaseSerializer):
    """
    Transform data to json string with json.dumps and json.loads to retrieve it back. Check
    https://docs.python.org/3/library/json.html#py-to-json-table for how types are converted.

    ujson will be used by default if available. Be careful with differences between built in
    json module and ujson:
        - ujson dumps supports bytes while json doesn't
        - ujson and json outputs may differ sometimes
    """

    def dumps(self, value):
        """
        Serialize the received value using ``json.dumps``.

        :param value: dict
        :returns: str
        """
        return json.dumps(value)

    def loads(self, value):
        """
        Deserialize value using ``json.loads``.

        :param value: str
        :returns: output of ``json.loads``.
        """
        if value is None:
            return None
        return json.loads(value)


class MsgPackSerializer(BaseSerializer):
    """
    Transform data to bytes using msgpack.dumps and msgpack.loads to retrieve it back. You need
    to have ``msgpack`` installed in order to be able to use this serializer.

    :param encoding: str. Can be used to change encoding param for ``msg.loads`` method.
        Default is utf-8.
    :param use_list: bool. Can be used to change use_list param for ``msgpack.loads`` method.
        Default is True.
    """

    def __init__(self, *args, use_list=True, **kwargs):
        if not msgpack:
            raise RuntimeError("msgpack not installed, MsgPackSerializer unavailable")
        self.use_list = use_list
        super().__init__(*args, **kwargs)

    def dumps(self, value):
        """
        Serialize the received value using ``msgpack.dumps``.

        :param value: obj
        :returns: bytes
        """
        return msgpack.dumps(value)

    def loads(self, value):
        """
        Deserialize value using ``msgpack.loads``.

        :param value: bytes
        :returns: obj
        """
        raw = False if self.encoding == "utf-8" else True
        if value is None:
            return None
        return msgpack.loads(value, raw=raw, use_list=self.use_list)



import random
import string
import asyncio

from marshmallow import fields, Schema, post_load

from aiocache import Cache
from aiocache.serializers import BaseSerializer


class RandomModel:
    MY_CONSTANT = "CONSTANT"

    def __init__(self, int_type=None, str_type=None, dict_type=None, list_type=None):
        self.int_type = int_type or random.randint(1, 10)
        self.str_type = str_type or random.choice(string.ascii_lowercase)
        self.dict_type = dict_type or {}
        self.list_type = list_type or []

    def __eq__(self, obj):
        return self.__dict__ == obj.__dict__


class MarshmallowSerializer(Schema, BaseSerializer):  # type: ignore[misc]
    int_type = fields.Integer()
    str_type = fields.String()
    dict_type = fields.Dict()
    list_type = fields.List(fields.Integer())

    # marshmallow Schema class doesn't play nicely with multiple inheritance and won't call
    # BaseSerializer.__init__
    encoding = 'utf-8'

    @post_load
    def build_my_type(self, data, **kwargs):
        return RandomModel(**data)

    class Meta:
        strict = True


cache = Cache(serializer=MarshmallowSerializer(), namespace="main")


async def serializer():
    model = RandomModel()
    await cache.set("key", model)

    result = await cache.get("key")

    assert result.int_type == model.int_type
    assert result.str_type == model.str_type
    assert result.dict_type == model.dict_type
    assert result.list_type == model.list_type


async def test_serializer():
    await serializer()
    await cache.delete("key")


if __name__ == "__main__":
    asyncio.run(test_serializer())



import asyncio
import zlib

from aiocache import Cache
from aiocache.serializers import BaseSerializer


class CompressionSerializer(BaseSerializer):

    # This is needed because zlib works with bytes.
    # this way the underlying backend knows how to
    # store/retrieve values
    DEFAULT_ENCODING = None

    def dumps(self, value):
        print("I've received:\n{}".format(value))
        compressed = zlib.compress(value.encode())
        print("But I'm storing:\n{}".format(compressed))
        return compressed

    def loads(self, value):
        print("I've retrieved:\n{}".format(value))
        decompressed = zlib.decompress(value).decode()
        print("But I'm returning:\n{}".format(decompressed))
        return decompressed


cache = Cache(Cache.REDIS, serializer=CompressionSerializer(), namespace="main")


async def serializer():
    text = (
        "Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt"
        "ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation"
        "ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in"
        "reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur"
        "sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit"
        "anim id est laborum.")
    await cache.set("key", text)
    print("-----------------------------------")
    real_value = await cache.get("key")
    compressed_value = await cache.raw("get", "main:key")
    assert len(compressed_value) < len(real_value.encode())


async def test_serializer():
    await serializer()
    await cache.delete("key")
    await cache.close()


if __name__ == "__main__":
    asyncio.run(test_serializer())



import asyncio
import logging
import uuid

from aiohttp import web

from aiocache import Cache

logging.getLogger("aiohttp.access").propagate = False


class CacheManager:
    def __init__(self, backend: str):
        backends = {
            "memory": Cache.MEMORY,
            "redis": Cache.REDIS,
            "memcached": Cache.MEMCACHED,
        }
        self.cache = Cache(backends[backend])

    async def get(self, key):
        return await self.cache.get(key, timeout=0.1)

    async def set(self, key, value):
        return await self.cache.set(key, value, timeout=0.1)

    async def close(self, *_):
        await self.cache.close()


async def handler_get(req):
    try:
        data = await req.app["cache"].get("testkey")
        if data:
            return web.Response(text=data)
    except asyncio.TimeoutError:
        return web.Response(status=404)

    data = str(uuid.uuid4())
    await req.app["cache"].set("testkey", data)
    return web.Response(text=str(data))


def run_server(backend: str) -> None:
    app = web.Application()
    app["cache"] = CacheManager(backend)
    app.on_shutdown.append(app["cache"].close)
    app.router.add_route("GET", "/", handler_get)
    web.run_app(app)



import asyncio
from unittest.mock import MagicMock

from aiocache.base import BaseCache


async def main():
    mocked_cache = MagicMock(spec=BaseCache)
    mocked_cache.get.return_value = "world"
    print(await mocked_cache.get("hello"))


if __name__ == "__main__":
    import sys

    if sys.version_info >= (3, 8):
        asyncio.run(main())



import sys
from contextlib import ExitStack
from unittest.mock import create_autospec, patch

import pytest

from aiocache import caches
from aiocache.base import BaseCache
from aiocache.plugins import BasePlugin

if sys.version_info < (3, 8):
    # Missing AsyncMock on 3.7
    collect_ignore_glob = ["*"]


@pytest.fixture(autouse=True)
def reset_caches():
    caches.set_config(
        {
            "default": {
                "cache": "aiocache.SimpleMemoryCache",
                "serializer": {"class": "aiocache.serializers.NullSerializer"},
            }
        }
    )


@pytest.fixture
def mock_cache(mocker):
    return create_autospec(BaseCache, instance=True)


@pytest.fixture
def mock_base_cache():
    """Return BaseCache instance with unimplemented methods mocked out."""
    plugin = create_autospec(BasePlugin, instance=True)
    cache = BaseCache(timeout=0.002, plugins=(plugin,))
    methods = ("_add", "_get", "_gets", "_set", "_multi_get", "_multi_set", "_delete",
               "_exists", "_increment", "_expire", "_clear", "_raw", "_close",
               "_redlock_release", "acquire_conn", "release_conn")
    with ExitStack() as stack:
        for f in methods:
            stack.enter_context(patch.object(cache, f, autospec=True))
        stack.enter_context(patch.object(cache, "_serializer", autospec=True))
        yield cache


@pytest.fixture
def base_cache():
    return BaseCache()


@pytest.fixture
async def redis_cache():
    from aiocache.backends.redis import RedisCache

    async with RedisCache() as cache:
        yield cache


@pytest.fixture
async def memcached_cache():
    from aiocache.backends.memcached import MemcachedCache

    async with MemcachedCache() as cache:
        yield cache



"""alt_key_builder.py

    ``key_builder`` is used in two contexts within ``aiocache``,
    with different meanings.
    1. Custom ``key_builder`` for a cache -- Prepends a namespace to the key
    2. Custom ``key_builder`` for a cache decorator -- Creates a cache key from
       the decorated callable and the callable's arguments

    --------------------------------------------------------------------------
    1. A custom ``key_builder`` for a cache can manipulate the name of a
       cache key; for example to meet naming requirements of the backend.

    ``key_builder`` can also optionally mark the key as belonging to a
    namespace group. This enables commonly used key names to be disambiguated
    by their ``namespace`` value. It also enables bulk operation on cache keys,
    such as expiring all keys in the same namespace.

    ``key_builder`` is expected (but not required) to prefix the passed key
    argument with the namespace argument. After initializing the cache object,
    the key builder can be accessed via the cache's ``build_key`` member.

    Args:
        key (str): undecorated key name
        namespace (str, optional): Prefix to add to the key. Defaults to None.

    Returns:
        By default, ``cache.build_key()`` returns ``f'{namespace}{sep}{key}'``,
        where some backends might include an optional separator, ``sep``.
        Some backends might strip or replace illegal characters, and encode
        the result before returning it. Typically str or bytes.

    --------------------------------------------------------------------------
    2. Custom ``key_builder`` for a cache decorator automatically generates a
       cache key from the call signature of the decorated callable. It does
       not accept a ``namespace`` parameter, and it should not add a
       naemspace to the key that it outputs.

    Args:
        func (callable): name of the decorated callable
        *args: Positional arguments when ``func`` was called.
        **kwargs: Keyword arguments when ``func`` was called.

    Returns (str):
        By default, the output key is a concatenation of the module and name
        of ``func`` + the positional arguments + the sorted keyword arguments.
"""
import asyncio
from typing import List, Dict

from aiocache import Cache, cached


async def demo_key_builders():
    await demo_cache_key_builders()
    await demo_cache_key_builders(namespace="demo")
    await demo_decorator_key_builders()


# 1. Custom ``key_builder`` for a cache
# -------------------------------------

def ensure_no_spaces(key, namespace=None, replace='_'):
    """Prefix key with namespace; replace each space with ``replace``"""
    aggregate_key = f"{namespace or ''}{key}"
    custom_key = aggregate_key.replace(' ', replace)
    return custom_key


def bytes_key(key, namespace=None):
    """Prefix key with namespace; convert output to bytes"""
    aggregate_key = f"{namespace or ''}{key}"
    custom_key = aggregate_key.encode()
    return custom_key


def fixed_key(key, namespace=None):
    """Ignore input, generate a fixed key"""
    unchanging_key = "universal key"
    return unchanging_key


async def demo_cache_key_builders(namespace=None):
    """Demonstrate usage and behavior of the custom key_builder functions"""
    cache_ns = "cache_namespace"
    async with Cache(Cache.MEMORY, key_builder=ensure_no_spaces, namespace=cache_ns) as cache:
        raw_key = "Key With Unwanted Spaces"
        return_value = 42
        await cache.add(raw_key, return_value, namespace=namespace)
        exists = await cache.exists(raw_key, namespace=namespace)
        assert exists is True
        custom_key = cache.build_key(raw_key, namespace=namespace)
        assert ' ' not in custom_key
        if namespace is not None:
            assert custom_key.startswith(namespace)
        else:
            # Using cache.namespace instead
            exists = await cache.exists(raw_key, namespace=cache_ns)
            assert exists is True
            custom_key = cache.build_key(raw_key, namespace=cache_ns)
            assert custom_key.startswith(cache_ns)
        cached_value = await cache.get(raw_key, namespace=namespace)
        assert cached_value == return_value
        await cache.delete(raw_key, namespace=namespace)

    async with Cache(Cache.MEMORY, key_builder=bytes_key) as cache:
        raw_key = "string-key"
        return_value = 42
        await cache.add(raw_key, return_value, namespace=namespace)
        exists = await cache.exists(raw_key, namespace=namespace)
        assert exists is True
        custom_key = cache.build_key(raw_key, namespace=namespace)
        assert isinstance(custom_key, bytes)
        cached_value = await cache.get(raw_key, namespace=namespace)
        assert cached_value == return_value
        await cache.delete(raw_key, namespace=namespace)

    async with Cache(Cache.MEMORY, key_builder=fixed_key) as cache:
        unchanging_key = "universal key"

        for raw_key, return_value in zip(
                ("key_1", "key_2", "key_3"),
                ("val_1", "val_2", "val_3")):
            await cache.set(raw_key, return_value, namespace=namespace)
            exists = await cache.exists(raw_key, namespace=namespace)
            assert exists is True
            custom_key = cache.build_key(raw_key, namespace=namespace)
            assert custom_key == unchanging_key
            cached_value = await cache.get(raw_key, namespace=namespace)
            assert cached_value == return_value

        # Cache key exists regardless of raw_key name
        for raw_key in ("key_1", "key_2", "key_3"):
            exists = await cache.exists(raw_key, namespace=namespace)
            assert exists is True

        cached_value = await cache.get(raw_key, namespace=namespace)
        assert cached_value == "val_3"  # The last value that was set
        await cache.delete(raw_key, namespace=namespace)

        # Deleting one cache key deletes them all
        for raw_key in ("key_1", "key_2", "key_3"):
            exists = await cache.exists(raw_key, namespace=namespace)
            assert exists is False


# 2. Custom ``key_builder`` for a cache decorator
# -----------------------------------------------

def ignore_kwargs(func, *args, **kwargs):
    """Do not use keyword arguments in the cache key's name"""
    return (
        (func.__module__ or "")
        + func.__name__
        + str(args)
    )


def module_override(func, *args, **kwargs):
    """Override the module-name prefix for the cache key"""
    ordered_kwargs = sorted(kwargs.items())
    return (
        "my_module_alias"
        + func.__name__
        + str(args)
        + str(ordered_kwargs)
    )


def hashed_args(*args, **kwargs):
    """Return a hashable key from a callable's parameters"""
    key = tuple()
    for arg in args:
        if isinstance(arg, List):
            key += tuple(hashed_args(_arg) for _arg in arg)
        elif isinstance(arg, Dict):
            key += tuple(sorted(
                (_key, hashed_args(_value)) for (_key, _value) in arg.items()
            ))
        else:
            key += (arg, )
    key += tuple(sorted(
        (_key, hashed_args(_value)) for (_key, _value) in kwargs.items()
    ))
    return key


def structured_key(func, *args, **kwargs):
    """String representation of a structured call signature"""
    key = tuple()
    key += (func.__module__ or '', )
    key += (func.__qualname__ or func.__name__, )
    key += hashed_args(*args, **kwargs)
    return str(key)


async def demo_decorator_key_builders():
    """Demonstrate usage and behavior of the custom key_builder functions"""
    await demo_ignore_kwargs_decorator()
    await demo_module_override_decorator()
    await demo_structured_key_decorator()


async def demo_ignore_kwargs_decorator():
    """Cache key from positional arguments in call to decorated function"""
    @cached(key_builder=ignore_kwargs)
    async def fn(a, b=2, c=3):
        return (a, b)

    (a, b) = (5, 1)
    demo_params = (
        dict(args=(a, b), kwargs=dict(c=3), ret=(a, b)),
        dict(args=(a, ), kwargs=dict(b=b, c=3), ret=(a, b)),
        dict(args=(a, ), kwargs=dict(c=3), ret=(a, b)),  # b from previous call
        dict(args=(a, b, 6), kwargs={}, ret=(a, b)),
    )
    demo_keys = list()

    for params in demo_params:
        args = params["args"]
        kwargs = params["kwargs"]

        await fn(*args, **kwargs)
        cache = fn.cache
        decorator = cached(key_builder=ignore_kwargs)
        key = decorator.get_cache_key(fn, args=args, kwargs=kwargs)
        exists = await cache.exists(key)
        assert exists is True
        assert key.endswith(str(args))
        cached_value = await cache.get(key)
        assert cached_value == params["ret"]
        demo_keys.append(key)

    assert demo_keys[1] == demo_keys[2]
    assert demo_keys[0] != demo_keys[1]
    assert demo_keys[0] != demo_keys[3]
    assert demo_keys[1] != demo_keys[3]

    for key in set(demo_keys):
        await cache.delete(key)


async def demo_module_override_decorator():
    """Cache key uses custom module name for decorated function"""
    @cached(key_builder=module_override)
    async def fn(a, b=2, c=3):
        return (a, b)

    (a, b) = (5, 1)
    args = (a, b)
    kwargs = dict(c=3)
    return_value = (a, b)

    await fn(*args, **kwargs)
    cache = fn.cache
    decorator = cached(key_builder=module_override)
    key = decorator.get_cache_key(fn, args=args, kwargs=kwargs)
    exists = await cache.exists(key)
    assert exists is True
    assert key.startswith("my_module_alias")
    cached_value = await cache.get(key)
    assert cached_value == return_value
    await cache.delete(key)


async def demo_structured_key_decorator():
    """Cache key expresses structure of decorated function call"""
    @cached(key_builder=structured_key)
    async def fn(a, b=2, c=3):
        return (a, b)

    (a, b) = (5, 1)
    args = (a, b)
    kwargs = dict(c=3)
    return_value = (a, b)
    fn_module = fn.__module__ or ''
    fn_name = fn.__qualname__ or fn.__name__
    key_name = str((fn_module, fn_name) + hashed_args(*args, **kwargs))

    await fn(*args, **kwargs)
    cache = fn.cache
    decorator = cached(key_builder=structured_key)
    key = decorator.get_cache_key(fn, args=args, kwargs=kwargs)
    exists = await cache.exists(key)
    assert exists is True
    assert key == key_name
    cached_value = await cache.get(key)
    assert cached_value == return_value
    await cache.delete(key)

# ---------------------------------------------------------------------------

if __name__ == "__main__":
    asyncio.run(demo_key_builders())



import tornado.web
import tornado.ioloop
from datetime import datetime
from aiocache import cached
from aiocache.serializers import JsonSerializer


class MainHandler(tornado.web.RequestHandler):

    # Due some incompatibilities between tornado and asyncio, caches can't use the "timeout" feature
    # in order to make it work, you will have to specify it always to 0
    @cached(key="my_custom_key", serializer=JsonSerializer(), timeout=0)
    async def time(self):
        return {"time": datetime.now().isoformat()}

    async def get(self):
        self.write(await self.time())


if __name__ == "__main__":
    tornado.ioloop.IOLoop.configure('tornado.platform.asyncio.AsyncIOLoop')
    app = tornado.web.Application([(r"/", MainHandler)])
    app.listen(8888)
    tornado.ioloop.IOLoop.current().start()



import asyncio
import random
import logging

from aiocache import Cache
from aiocache.plugins import HitMissRatioPlugin, TimingPlugin, BasePlugin


logger = logging.getLogger(__name__)


class MyCustomPlugin(BasePlugin):

    async def pre_set(self, *args, **kwargs):
        logger.info("I'm the pre_set hook being called with %s %s" % (args, kwargs))

    async def post_set(self, *args, **kwargs):
        logger.info("I'm the post_set hook being called with %s %s" % (args, kwargs))


cache = Cache(
    plugins=[HitMissRatioPlugin(), TimingPlugin(), MyCustomPlugin()],
    namespace="main")


async def run():
    await cache.set("a", "1")
    await cache.set("b", "2")
    await cache.set("c", "3")
    await cache.set("d", "4")

    possible_keys = ["a", "b", "c", "d", "e", "f"]

    for t in range(1000):
        await cache.get(random.choice(possible_keys))

    assert cache.hit_miss_ratio["hit_ratio"] > 0.5
    assert cache.hit_miss_ratio["total"] == 1000

    assert cache.profiling["get_min"] > 0
    assert cache.profiling["set_min"] > 0
    assert cache.profiling["get_max"] > 0
    assert cache.profiling["set_max"] > 0

    print(cache.hit_miss_ratio)
    print(cache.profiling)


async def test_run():
    await run()
    await cache.delete("a")
    await cache.delete("b")
    await cache.delete("c")
    await cache.delete("d")


if __name__ == "__main__":
    asyncio.run(test_run())



from unittest.mock import Mock, patch

import pytest

from aiocache import AIOCACHE_CACHES, Cache, caches
from aiocache.backends.memory import SimpleMemoryCache
from aiocache.exceptions import InvalidCacheType
from aiocache.factory import _class_from_string, _create_cache
from aiocache.plugins import HitMissRatioPlugin, TimingPlugin
from aiocache.serializers import JsonSerializer, PickleSerializer


CACHE_NAMES = [Cache.MEMORY.NAME]

try:
    from aiocache.backends.memcached import MemcachedCache
except ImportError:
    MemcachedCache = None
else:
    assert Cache.MEMCACHED is not None
    CACHE_NAMES.append(Cache.MEMCACHED.NAME)

try:
    from aiocache.backends.redis import RedisCache
except ImportError:
    RedisCache = None
else:
    assert Cache.REDIS is not None
    CACHE_NAMES.append(Cache.REDIS.NAME)


@pytest.mark.redis
def test_class_from_string():
    assert _class_from_string("aiocache.RedisCache") == RedisCache


@pytest.mark.redis
def test_create_simple_cache():
    redis = _create_cache(RedisCache, endpoint="127.0.0.10", port=6378)

    assert isinstance(redis, RedisCache)
    assert redis.endpoint == "127.0.0.10"
    assert redis.port == 6378


def test_create_cache_with_everything():
    cache = _create_cache(
        SimpleMemoryCache,
        serializer={"class": PickleSerializer, "encoding": "encoding"},
        plugins=[{"class": "aiocache.plugins.TimingPlugin"}],
    )

    assert isinstance(cache.serializer, PickleSerializer)
    assert cache.serializer.encoding == "encoding"
    assert isinstance(cache.plugins[0], TimingPlugin)


class TestCache:
    def test_cache_types(self):
        assert Cache.MEMORY == SimpleMemoryCache
        assert Cache.REDIS == RedisCache
        assert Cache.MEMCACHED == MemcachedCache

    @pytest.mark.parametrize("cache_type", CACHE_NAMES)
    async def test_new(self, cache_type):
        kwargs = {"a": 1, "b": 2}
        cache_class = Cache.get_scheme_class(cache_type)

        with patch("aiocache.{}.__init__".format(cache_class.__name__)) as init:
            cache = Cache(cache_class, **kwargs)
            assert isinstance(cache, cache_class)
            init.assert_called_once_with(**kwargs)

    def test_new_defaults_to_memory(self):
        assert isinstance(Cache(), Cache.MEMORY)

    def test_new_invalid_cache_raises(self):
        with pytest.raises(InvalidCacheType) as e:
            Cache(object)
        assert str(e.value) == "Invalid cache type, you can only use {}".format(
            list(AIOCACHE_CACHES.keys())
        )

    @pytest.mark.parametrize("scheme", CACHE_NAMES)
    def test_get_scheme_class(self, scheme):
        assert Cache.get_scheme_class(scheme) == AIOCACHE_CACHES[scheme]

    def test_get_scheme_class_invalid(self):
        with pytest.raises(InvalidCacheType):
            Cache.get_scheme_class("http")

    @pytest.mark.parametrize("scheme", CACHE_NAMES)
    def test_from_url_returns_cache_from_scheme(self, scheme):
        assert isinstance(Cache.from_url("{}://".format(scheme)), Cache.get_scheme_class(scheme))

    @pytest.mark.parametrize(
        "url,expected_args",
        [
            ("redis://", {}),
            ("redis://localhost", {"endpoint": "localhost"}),
            ("redis://localhost/", {"endpoint": "localhost"}),
            ("redis://localhost:6379", {"endpoint": "localhost", "port": 6379}),
            (
                "redis://localhost/?arg1=arg1&arg2=arg2",
                {"endpoint": "localhost", "arg1": "arg1", "arg2": "arg2"},
            ),
            (
                "redis://localhost:6379/?arg1=arg1&arg2=arg2",
                {"endpoint": "localhost", "port": 6379, "arg1": "arg1", "arg2": "arg2"},
            ),
            ("redis:///?arg1=arg1", {"arg1": "arg1"}),
            ("redis:///?arg2=arg2", {"arg2": "arg2"}),
            (
                "redis://:password@localhost:6379",
                {"endpoint": "localhost", "password": "password", "port": 6379},
            ),
            (
                "redis://:password@localhost:6379?password=pass",
                {"endpoint": "localhost", "password": "password", "port": 6379},
            ),
        ],
    )
    def test_from_url_calls_cache_with_args(self, url, expected_args):
        with patch("aiocache.factory.Cache", autospec=True) as mock:
            Cache.from_url(url)

        mock.assert_called_once_with(mock.get_scheme_class.return_value, **expected_args)

    def test_calls_parse_uri_path_from_cache(self):
        p_mock = Mock(spec_set=(), return_value={"arg1": "arg1"})
        with patch("aiocache.factory.Cache", autospec=True) as mock:
            mock.get_scheme_class.return_value.parse_uri_path = p_mock
            Cache.from_url("redis:///")

        mock.get_scheme_class.return_value.parse_uri_path.assert_called_once_with("/")
        mock.assert_called_once_with(mock.get_scheme_class.return_value, arg1="arg1")

    def test_from_url_invalid_protocol(self):
        with pytest.raises(InvalidCacheType):
            Cache.from_url("http://")


class TestCacheHandler:
    @pytest.fixture(autouse=True)
    def remove_caches(self):
        caches._caches = {}

    def test_add_new_entry(self):
        alias = "memory"
        config = {
            "cache": "aiocache.SimpleMemoryCache",
            "serializer": {"class": "aiocache.serializers.StringSerializer"},
        }
        caches.add(alias, config)

        assert caches.get_config()[alias] == config

    def test_add_updates_existing_entry(self):
        alias = "memory"
        config = {
            "cache": "aiocache.SimpleMemoryCache",
            "serializer": {"class": "aiocache.serializers.StringSerializer"},
        }
        caches.add(alias, {})
        caches.add(alias, config)

        assert caches.get_config()[alias] == config

    def test_get_wrong_alias(self):
        with pytest.raises(KeyError):
            caches.get("wrong_cache")

        with pytest.raises(KeyError):
            caches.create("wrong_cache")

    def test_reuse_instance(self):
        assert caches.get("default") is caches.get("default")

    def test_create_not_reuse(self):
        assert caches.create("default") is not caches.create("default")

    @pytest.mark.redis
    def test_create_extra_args(self):
        caches.set_config(
            {
                "default": {
                    "cache": "aiocache.RedisCache",
                    "endpoint": "127.0.0.9",
                    "db": 10,
                    "port": 6378,
                }
            }
        )
        cache = caches.create("default", namespace="whatever", endpoint="127.0.0.10", db=10)
        assert cache.namespace == "whatever"
        assert cache.endpoint == "127.0.0.10"
        assert cache.db == 10

    @pytest.mark.redis
    def test_retrieve_cache(self):
        caches.set_config(
            {
                "default": {
                    "cache": "aiocache.RedisCache",
                    "endpoint": "127.0.0.10",
                    "port": 6378,
                    "ttl": 10,
                    "serializer": {
                        "class": "aiocache.serializers.PickleSerializer",
                        "encoding": "encoding",
                    },
                    "plugins": [
                        {"class": "aiocache.plugins.HitMissRatioPlugin"},
                        {"class": "aiocache.plugins.TimingPlugin"},
                    ],
                }
            }
        )

        cache = caches.get("default")
        assert isinstance(cache, RedisCache)
        assert cache.endpoint == "127.0.0.10"
        assert cache.port == 6378
        assert cache.ttl == 10
        assert isinstance(cache.serializer, PickleSerializer)
        assert cache.serializer.encoding == "encoding"
        assert len(cache.plugins) == 2

    @pytest.mark.redis
    def test_retrieve_cache_new_instance(self):
        caches.set_config(
            {
                "default": {
                    "cache": "aiocache.RedisCache",
                    "endpoint": "127.0.0.10",
                    "port": 6378,
                    "serializer": {
                        "class": "aiocache.serializers.PickleSerializer",
                        "encoding": "encoding",
                    },
                    "plugins": [
                        {"class": "aiocache.plugins.HitMissRatioPlugin"},
                        {"class": "aiocache.plugins.TimingPlugin"},
                    ],
                }
            }
        )

        cache = caches.create("default")
        assert isinstance(cache, RedisCache)
        assert cache.endpoint == "127.0.0.10"
        assert cache.port == 6378
        assert isinstance(cache.serializer, PickleSerializer)
        assert cache.serializer.encoding == "encoding"
        assert len(cache.plugins) == 2

    @pytest.mark.redis
    def test_multiple_caches(self):
        caches.set_config(
            {
                "default": {
                    "cache": "aiocache.RedisCache",
                    "endpoint": "127.0.0.10",
                    "port": 6378,
                    "serializer": {"class": "aiocache.serializers.PickleSerializer"},
                    "plugins": [
                        {"class": "aiocache.plugins.HitMissRatioPlugin"},
                        {"class": "aiocache.plugins.TimingPlugin"},
                    ],
                },
                "alt": {"cache": "aiocache.SimpleMemoryCache"},
            }
        )

        default = caches.get("default")
        alt = caches.get("alt")

        assert isinstance(default, RedisCache)
        assert default.endpoint == "127.0.0.10"
        assert default.port == 6378
        assert isinstance(default.serializer, PickleSerializer)
        assert len(default.plugins) == 2

        assert isinstance(alt, SimpleMemoryCache)

    def test_default_caches(self):
        assert caches.get_config() == {
            "default": {
                "cache": "aiocache.SimpleMemoryCache",
                "serializer": {"class": "aiocache.serializers.NullSerializer"},
            }
        }

    def test_get_alias_config(self):
        assert caches.get_alias_config("default") == {
            "cache": "aiocache.SimpleMemoryCache",
            "serializer": {"class": "aiocache.serializers.NullSerializer"},
        }

    def test_set_empty_config(self):
        with pytest.raises(ValueError):
            caches.set_config({})

    def test_set_config_updates_existing_values(self):
        assert not isinstance(caches.get("default").serializer, JsonSerializer)
        caches.set_config(
            {
                "default": {
                    "cache": "aiocache.SimpleMemoryCache",
                    "serializer": {"class": "aiocache.serializers.JsonSerializer"},
                }
            }
        )
        assert isinstance(caches.get("default").serializer, JsonSerializer)

    def test_set_config_removes_existing_caches(self):
        caches.set_config(
            {
                "default": {"cache": "aiocache.SimpleMemoryCache"},
                "alt": {"cache": "aiocache.SimpleMemoryCache"},
            }
        )
        caches.get("default")
        caches.get("alt")
        assert len(caches._caches) == 2

        caches.set_config(
            {
                "default": {"cache": "aiocache.SimpleMemoryCache"},
                "alt": {"cache": "aiocache.SimpleMemoryCache"},
            }
        )
        assert caches._caches == {}

    def test_set_config_no_default(self):
        with pytest.raises(ValueError):
            caches.set_config(
                {
                    "no_default": {
                        "cache": "aiocache.RedisCache",
                        "endpoint": "127.0.0.10",
                        "port": 6378,
                        "serializer": {"class": "aiocache.serializers.PickleSerializer"},
                        "plugins": [
                            {"class": "aiocache.plugins.HitMissRatioPlugin"},
                            {"class": "aiocache.plugins.TimingPlugin"},
                        ],
                    }
                }
            )

    @pytest.mark.redis
    def test_ensure_plugins_order(self):
        caches.set_config(
            {
                "default": {
                    "cache": "aiocache.RedisCache",
                    "plugins": [
                        {"class": "aiocache.plugins.HitMissRatioPlugin"},
                        {"class": "aiocache.plugins.TimingPlugin"},
                    ],
                }
            }
        )

        cache = caches.get("default")
        assert isinstance(cache.plugins[0], HitMissRatioPlugin)

        cache = caches.create("default")
        assert isinstance(cache.plugins[0], HitMissRatioPlugin)



import itertools
import warnings

import redis.asyncio as redis
from redis.exceptions import ResponseError as IncrbyException

from aiocache.base import BaseCache, _ensure_key
from aiocache.serializers import JsonSerializer


_NOT_SET = object()


class RedisBackend(BaseCache):
    RELEASE_SCRIPT = (
        "if redis.call('get',KEYS[1]) == ARGV[1] then"
        " return redis.call('del',KEYS[1])"
        " else"
        " return 0"
        " end"
    )

    CAS_SCRIPT = (
        "if redis.call('get',KEYS[1]) == ARGV[2] then"
        "  if #ARGV == 4 then"
        "   return redis.call('set', KEYS[1], ARGV[1], ARGV[3], ARGV[4])"
        "  else"
        "   return redis.call('set', KEYS[1], ARGV[1])"
        "  end"
        " else"
        " return 0"
        " end"
    )

    def __init__(
        self,
        endpoint="127.0.0.1",
        port=6379,
        db=0,
        password=None,
        pool_min_size=_NOT_SET,
        pool_max_size=None,
        create_connection_timeout=None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        if pool_min_size is not _NOT_SET:
            warnings.warn(
                "Parameter 'pool_min_size' is deprecated since aiocache 0.12",
                DeprecationWarning,
            )

        self.endpoint = endpoint
        self.port = int(port)
        self.db = int(db)
        self.password = password
        # TODO: Remove int() call some time after adding type annotations.
        self.pool_max_size = None if pool_max_size is None else int(pool_max_size)
        self.create_connection_timeout = (
            float(create_connection_timeout) if create_connection_timeout else None
        )

        # NOTE: decoding can't be controlled on API level after switching to
        # redis, we need to disable decoding on global/connection level
        # (decode_responses=False), because some of the values are saved as
        # bytes directly, like pickle serialized values, which may raise an
        # exception when decoded with 'utf-8'.
        self.client = redis.Redis(host=self.endpoint, port=self.port, db=self.db,
                                  password=self.password, decode_responses=False,
                                  socket_connect_timeout=self.create_connection_timeout,
                                  max_connections=self.pool_max_size)

    async def _get(self, key, encoding="utf-8", _conn=None):
        value = await self.client.get(key)
        if encoding is None or value is None:
            return value
        return value.decode(encoding)

    async def _gets(self, key, encoding="utf-8", _conn=None):
        return await self._get(key, encoding=encoding, _conn=_conn)

    async def _multi_get(self, keys, encoding="utf-8", _conn=None):
        values = await self.client.mget(*keys)
        if encoding is None:
            return values
        return [v if v is None else v.decode(encoding) for v in values]

    async def _set(self, key, value, ttl=None, _cas_token=None, _conn=None):
        if _cas_token is not None:
            return await self._cas(key, value, _cas_token, ttl=ttl, _conn=_conn)
        if ttl is None:
            return await self.client.set(key, value)
        if isinstance(ttl, float):
            ttl = int(ttl * 1000)
            return await self.client.psetex(key, ttl, value)
        return await self.client.setex(key, ttl, value)

    async def _cas(self, key, value, token, ttl=None, _conn=None):
        args = ()
        if ttl is not None:
            args = ("PX", int(ttl * 1000)) if isinstance(ttl, float) else ("EX", ttl)
        return await self._raw("eval", self.CAS_SCRIPT, 1, key, value, token, *args, _conn=_conn)

    async def _multi_set(self, pairs, ttl=None, _conn=None):
        ttl = ttl or 0

        flattened = list(itertools.chain.from_iterable((key, value) for key, value in pairs))

        if ttl:
            await self.__multi_set_ttl(flattened, ttl)
        else:
            await self.client.execute_command("MSET", *flattened)

        return True

    async def __multi_set_ttl(self, flattened, ttl):
        async with self.client.pipeline(transaction=True) as p:
            p.execute_command("MSET", *flattened)
            ttl, exp = (int(ttl * 1000), p.pexpire) if isinstance(ttl, float) else (ttl, p.expire)
            for key in flattened[::2]:
                exp(key, time=ttl)
            await p.execute()

    async def _add(self, key, value, ttl=None, _conn=None):
        kwargs = {"nx": True}
        if isinstance(ttl, float):
            kwargs["px"] = int(ttl * 1000)
        else:
            kwargs["ex"] = ttl
        was_set = await self.client.set(key, value, **kwargs)
        if not was_set:
            raise ValueError("Key {} already exists, use .set to update the value".format(key))
        return was_set

    async def _exists(self, key, _conn=None):
        number = await self.client.exists(key)
        return bool(number)

    async def _increment(self, key, delta, _conn=None):
        try:
            return await self.client.incrby(key, delta)
        except IncrbyException:
            raise TypeError("Value is not an integer") from None

    async def _expire(self, key, ttl, _conn=None):
        if ttl == 0:
            return await self.client.persist(key)
        return await self.client.expire(key, ttl)

    async def _delete(self, key, _conn=None):
        return await self.client.delete(key)

    async def _clear(self, namespace=None, _conn=None):
        if namespace:
            keys = await self.client.keys("{}:*".format(namespace))
            if keys:
                await self.client.delete(*keys)
        else:
            await self.client.flushdb()
        return True

    async def _raw(self, command, *args, encoding="utf-8", _conn=None, **kwargs):
        value = await getattr(self.client, command)(*args, **kwargs)
        if encoding is not None:
            if command == "get" and value is not None:
                value = value.decode(encoding)
            elif command in {"keys", "mget"}:
                value = [v if v is None else v.decode(encoding) for v in value]
        return value

    async def _redlock_release(self, key, value):
        return await self._raw("eval", self.RELEASE_SCRIPT, 1, key, value)

    async def _close(self, *args, _conn=None, **kwargs):
        await self.client.close()


class RedisCache(RedisBackend):
    """
    Redis cache implementation with the following components as defaults:
        - serializer: :class:`aiocache.serializers.JsonSerializer`
        - plugins: []

    Config options are:

    :param serializer: obj derived from :class:`aiocache.serializers.BaseSerializer`.
    :param plugins: list of :class:`aiocache.plugins.BasePlugin` derived classes.
    :param namespace: string to use as default prefix for the key used in all operations of
        the backend. Default is None.
    :param timeout: int or float in seconds specifying maximum timeout for the operations to last.
        By default its 5.
    :param endpoint: str with the endpoint to connect to. Default is "127.0.0.1".
    :param port: int with the port to connect to. Default is 6379.
    :param db: int indicating database to use. Default is 0.
    :param password: str indicating password to use. Default is None.
    :param pool_max_size: int maximum pool size for the redis connections pool. Default is None.
    :param create_connection_timeout: int timeout for the creation of connection. Default is None
    """

    NAME = "redis"

    def __init__(self, serializer=None, **kwargs):
        super().__init__(serializer=serializer or JsonSerializer(), **kwargs)

    @classmethod
    def parse_uri_path(cls, path):
        """
        Given a uri path, return the Redis specific configuration
        options in that path string according to iana definition
        http://www.iana.org/assignments/uri-schemes/prov/redis

        :param path: string containing the path. Example: "/0"
        :return: mapping containing the options. Example: {"db": "0"}
        """
        options = {}
        db, *_ = path[1:].split("/")
        if db:
            options["db"] = db
        return options

    def _build_key(self, key, namespace=None):
        if namespace is not None:
            return "{}{}{}".format(
                namespace, ":" if namespace else "", _ensure_key(key))
        if self.namespace is not None:
            return "{}{}{}".format(
                self.namespace, ":" if self.namespace else "", _ensure_key(key))
        return key

    def __repr__(self):  # pragma: no cover
        return "RedisCache ({}:{})".format(self.endpoint, self.port)



from unittest.mock import AsyncMock, patch

import aiomcache
import pytest

from aiocache.backends.memcached import MemcachedBackend, MemcachedCache
from aiocache.base import BaseCache, _ensure_key
from aiocache.serializers import JsonSerializer
from ...utils import Keys


@pytest.fixture
def memcached():
    memcached = MemcachedBackend()
    with patch.object(memcached, "client", autospec=True) as m:
        # Autospec messes up the signature on the decorated methods.
        for method in (
            "get", "gets", "multi_get", "stats", "set", "cas", "replace",
            "append", "prepend", "incr", "decr", "touch", "version", "flush_all"
        ):
            setattr(m, method, AsyncMock(return_value=None, spec_set=()))
        m.add = AsyncMock(return_value=True, spec_set=())
        m.delete = AsyncMock(return_value=True, spec_set=())

        yield memcached


class TestMemcachedBackend:
    def test_setup(self):
        with patch.object(aiomcache, "Client", autospec=True) as aiomcache_client:
            memcached = MemcachedBackend()

            aiomcache_client.assert_called_with("127.0.0.1", 11211, pool_size=2)

        assert memcached.endpoint == "127.0.0.1"
        assert memcached.port == 11211
        assert memcached.pool_size == 2

    def test_setup_override(self):
        with patch.object(aiomcache, "Client", autospec=True) as aiomcache_client:
            memcached = MemcachedBackend(endpoint="127.0.0.2", port=2, pool_size=10)

            aiomcache_client.assert_called_with("127.0.0.2", 2, pool_size=10)

        assert memcached.endpoint == "127.0.0.2"
        assert memcached.port == 2
        assert memcached.pool_size == 10

    def test_setup_casts(self):
        with patch.object(aiomcache, "Client", autospec=True) as aiomcache_client:
            memcached = MemcachedBackend(pool_size="10")

            aiomcache_client.assert_called_with("127.0.0.1", 11211, pool_size=10)

        assert memcached.pool_size == 10

    async def test_get(self, memcached):
        memcached.client.get.return_value = b"value"
        assert await memcached._get(Keys.KEY) == "value"
        memcached.client.get.assert_called_with(Keys.KEY)

    async def test_gets(self, memcached):
        memcached.client.gets.return_value = b"value", 12345
        assert await memcached._gets(Keys.KEY) == 12345
        memcached.client.gets.assert_called_with(Keys.KEY.encode())

    async def test_get_none(self, memcached):
        memcached.client.get.return_value = None
        assert await memcached._get(Keys.KEY) is None
        memcached.client.get.assert_called_with(Keys.KEY)

    async def test_get_no_encoding(self, memcached):
        memcached.client.get.return_value = b"value"
        assert await memcached._get(Keys.KEY, encoding=None) == b"value"
        memcached.client.get.assert_called_with(Keys.KEY)

    async def test_set(self, memcached):
        await memcached._set(Keys.KEY, "value")
        memcached.client.set.assert_called_with(Keys.KEY, b"value", exptime=0)

        await memcached._set(Keys.KEY, "value", ttl=1)
        memcached.client.set.assert_called_with(Keys.KEY, b"value", exptime=1)

    async def test_set_float_ttl(self, memcached):
        memcached.client.set.side_effect = aiomcache.exceptions.ValidationException("msg")
        with pytest.raises(TypeError) as exc_info:
            await memcached._set(Keys.KEY, "value", ttl=0.1)
        assert str(exc_info.value) == "aiomcache error: msg"

    async def test_set_cas_token(self, mocker, memcached):
        mocker.spy(memcached, "_cas")
        await memcached._set(Keys.KEY, "value", _cas_token="token")
        memcached._cas.assert_called_with(Keys.KEY, b"value", "token", ttl=0, _conn=None)

    async def test_cas(self, memcached):
        memcached.client.cas.return_value = True
        assert await memcached._cas(Keys.KEY, b"value", "token", ttl=0) is True
        memcached.client.cas.assert_called_with(Keys.KEY, b"value", "token", exptime=0)

    async def test_cas_fail(self, memcached):
        memcached.client.cas.return_value = False
        assert await memcached._cas(Keys.KEY, b"value", "token", ttl=0) is False
        memcached.client.cas.assert_called_with(Keys.KEY, b"value", "token", exptime=0)

    async def test_multi_get(self, memcached):
        memcached.client.multi_get.return_value = [b"value", b"random"]
        assert await memcached._multi_get([Keys.KEY, Keys.KEY_1]) == ["value", "random"]
        memcached.client.multi_get.assert_called_with(Keys.KEY, Keys.KEY_1)

    async def test_multi_get_none(self, memcached):
        memcached.client.multi_get.return_value = [b"value", None]
        assert await memcached._multi_get([Keys.KEY, Keys.KEY_1]) == ["value", None]
        memcached.client.multi_get.assert_called_with(Keys.KEY, Keys.KEY_1)

    async def test_multi_get_no_encoding(self, memcached):
        memcached.client.multi_get.return_value = [b"value", None]
        assert await memcached._multi_get([Keys.KEY, Keys.KEY_1], encoding=None) == [
            b"value",
            None,
        ]
        memcached.client.multi_get.assert_called_with(Keys.KEY, Keys.KEY_1)

    async def test_multi_set(self, memcached):
        await memcached._multi_set([(Keys.KEY, "value"), (Keys.KEY_1, "random")])
        memcached.client.set.assert_any_call(Keys.KEY, b"value", exptime=0)
        memcached.client.set.assert_any_call(Keys.KEY_1, b"random", exptime=0)
        assert memcached.client.set.call_count == 2

        await memcached._multi_set([(Keys.KEY, "value"), (Keys.KEY_1, "random")], ttl=1)
        memcached.client.set.assert_any_call(Keys.KEY, b"value", exptime=1)
        memcached.client.set.assert_any_call(Keys.KEY_1, b"random", exptime=1)
        assert memcached.client.set.call_count == 4

    async def test_multi_set_float_ttl(self, memcached):
        memcached.client.set.side_effect = aiomcache.exceptions.ValidationException("msg")
        with pytest.raises(TypeError) as exc_info:
            await memcached._multi_set([(Keys.KEY, "value"), (Keys.KEY_1, "random")], ttl=0.1)
        assert str(exc_info.value) == "aiomcache error: msg"

    async def test_add(self, memcached):
        await memcached._add(Keys.KEY, "value")
        memcached.client.add.assert_called_with(Keys.KEY, b"value", exptime=0)

        await memcached._add(Keys.KEY, "value", ttl=1)
        memcached.client.add.assert_called_with(Keys.KEY, b"value", exptime=1)

    async def test_add_existing(self, memcached):
        memcached.client.add.return_value = False
        with pytest.raises(ValueError):
            await memcached._add(Keys.KEY, "value")

    async def test_add_float_ttl(self, memcached):
        memcached.client.add.side_effect = aiomcache.exceptions.ValidationException("msg")
        with pytest.raises(TypeError) as exc_info:
            await memcached._add(Keys.KEY, "value", 0.1)
        assert str(exc_info.value) == "aiomcache error: msg"

    async def test_exists(self, memcached):
        await memcached._exists(Keys.KEY)
        memcached.client.append.assert_called_with(Keys.KEY, b"")

    async def test_increment(self, memcached):
        await memcached._increment(Keys.KEY, 2)
        memcached.client.incr.assert_called_with(Keys.KEY, 2)

    async def test_increment_negative(self, memcached):
        await memcached._increment(Keys.KEY, -2)
        memcached.client.decr.assert_called_with(Keys.KEY, 2)

    async def test_increment_missing(self, memcached):
        memcached.client.incr.side_effect = aiomcache.exceptions.ClientException("NOT_FOUND")
        await memcached._increment(Keys.KEY, 2)
        memcached.client.incr.assert_called_with(Keys.KEY, 2)
        memcached.client.set.assert_called_with(Keys.KEY, b"2", exptime=0)

    async def test_increment_missing_negative(self, memcached):
        memcached.client.decr.side_effect = aiomcache.exceptions.ClientException("NOT_FOUND")
        await memcached._increment(Keys.KEY, -2)
        memcached.client.decr.assert_called_with(Keys.KEY, 2)
        memcached.client.set.assert_called_with(Keys.KEY, b"-2", exptime=0)

    async def test_increment_typerror(self, memcached):
        memcached.client.incr.side_effect = aiomcache.exceptions.ClientException("msg")
        with pytest.raises(TypeError) as exc_info:
            await memcached._increment(Keys.KEY, 2)
        assert str(exc_info.value) == "aiomcache error: msg"

    async def test_expire(self, memcached):
        await memcached._expire(Keys.KEY, 1)
        memcached.client.touch.assert_called_with(Keys.KEY, 1)

    async def test_delete(self, memcached):
        assert await memcached._delete(Keys.KEY) == 1
        memcached.client.delete.assert_called_with(Keys.KEY)

    async def test_delete_missing(self, memcached):
        memcached.client.delete.return_value = False
        assert await memcached._delete(Keys.KEY) == 0
        memcached.client.delete.assert_called_with(Keys.KEY)

    async def test_clear(self, memcached):
        await memcached._clear()
        memcached.client.flush_all.assert_called_with()

    async def test_clear_with_namespace(self, memcached):
        with pytest.raises(ValueError):
            await memcached._clear("nm")

    async def test_raw(self, memcached):
        await memcached._raw("get", Keys.KEY)
        await memcached._raw("set", Keys.KEY, 1)
        memcached.client.get.assert_called_with(Keys.KEY)
        memcached.client.set.assert_called_with(Keys.KEY, 1)

    async def test_raw_bytes(self, memcached):
        await memcached._raw("set", Keys.KEY, "asd")
        await memcached._raw("get", Keys.KEY, encoding=None)
        memcached.client.get.assert_called_with(Keys.KEY)
        memcached.client.set.assert_called_with(Keys.KEY, "asd")

    async def test_redlock_release(self, mocker, memcached):
        mocker.spy(memcached, "_delete")
        await memcached._redlock_release(Keys.KEY, "random")
        memcached._delete.assert_called_with(Keys.KEY)

    async def test_close(self, memcached):
        await memcached._close()
        assert memcached.client.close.call_count == 1


class TestMemcachedCache:
    @pytest.fixture
    def set_test_namespace(self, memcached_cache):
        memcached_cache.namespace = "test"
        yield
        memcached_cache.namespace = None

    def test_name(self):
        assert MemcachedCache.NAME == "memcached"

    def test_inheritance(self):
        assert isinstance(MemcachedCache(), BaseCache)

    def test_default_serializer(self):
        assert isinstance(MemcachedCache().serializer, JsonSerializer)

    def test_parse_uri_path(self):
        assert MemcachedCache().parse_uri_path("/1/2/3") == {}

    @pytest.mark.parametrize(
        "namespace, expected",
        ([None, "test" + _ensure_key(Keys.KEY)], ["", _ensure_key(Keys.KEY)], ["my_ns", "my_ns" + _ensure_key(Keys.KEY)]),  # type: ignore[attr-defined]  # noqa: B950
    )
    def test_build_key_bytes(self, set_test_namespace, memcached_cache, namespace, expected):
        assert memcached_cache.build_key(Keys.KEY, namespace=namespace) == expected.encode()

    def test_build_key_no_namespace(self, memcached_cache):
        assert memcached_cache.build_key(Keys.KEY, namespace=None) == Keys.KEY.encode()

    def test_build_key_no_spaces(self, memcached_cache):
        assert memcached_cache.build_key("hello world") == b"hello_world"



import asyncio

import pytest

from aiocache.lock import OptimisticLock, OptimisticLockError, RedLock
from aiocache.serializers import StringSerializer
from ..utils import KEY_LOCK, Keys


@pytest.fixture
def lock(cache):
    return RedLock(cache, Keys.KEY, 20)


def build_key(key, namespace=None):
    return "custom_key"


def build_key_bytes(key, namespace=None):
    return b"custom_key"


@pytest.fixture
def custom_redis_cache(mocker, redis_cache, build_key=build_key):
    mocker.patch.object(redis_cache, "build_key", new=build_key)
    yield redis_cache


@pytest.fixture
def custom_memory_cache(mocker, memory_cache, build_key=build_key):
    mocker.patch.object(memory_cache, "build_key", new=build_key)
    yield memory_cache


@pytest.fixture
def custom_memcached_cache(mocker, memcached_cache, build_key=build_key_bytes):
    mocker.patch.object(memcached_cache, "build_key", new=build_key)
    yield memcached_cache


class TestRedLock:
    async def test_acquire(self, cache, lock):
        cache.serializer = StringSerializer()
        async with lock:
            assert await cache.get(KEY_LOCK) == lock._value

    async def test_release_does_nothing_when_no_lock(self, lock):
        assert await lock.__aexit__("exc_type", "exc_value", "traceback") is None

    async def test_acquire_release(self, cache, lock):
        async with lock:
            pass
        assert await cache.get(KEY_LOCK) is None

    async def test_locking_dogpile(self, mocker, cache):
        mocker.spy(cache, "get")
        mocker.spy(cache, "set")
        mocker.spy(cache, "_add")

        async def dummy():
            res = await cache.get(Keys.KEY)
            assert res is None

            async with RedLock(cache, Keys.KEY, lease=5):
                res = await cache.get(Keys.KEY)
                if res is not None:
                    return
                await asyncio.sleep(0.1)
                await cache.set(Keys.KEY, "value")

        await asyncio.gather(dummy(), dummy(), dummy(), dummy())
        assert cache._add.call_count == 4
        assert cache.get.call_count == 8
        assert cache.set.call_count == 1, cache.set.call_args_list

    async def test_locking_dogpile_lease_expiration(self, cache):
        async def dummy() -> None:
            res = await cache.get(Keys.KEY)
            assert res is None

            # Lease should expire before cache is set, so res is still None.
            async with RedLock(cache, Keys.KEY, lease=1):
                res = await cache.get(Keys.KEY)
                assert res is None
                await asyncio.sleep(1.1)
                await cache.set(Keys.KEY, "value")

        await asyncio.gather(dummy(), dummy(), dummy(), dummy())

    async def test_locking_dogpile_propagates_exceptions(self, cache):
        async def dummy():
            async with RedLock(cache, Keys.KEY, lease=1):
                raise ValueError()

        with pytest.raises(ValueError):
            await dummy()


class TestMemoryRedLock:
    @pytest.fixture
    def lock(self, memory_cache):
        return RedLock(memory_cache, Keys.KEY, 20)

    async def test_acquire_key_builder(self, custom_memory_cache, lock):
        async with lock:
            assert await custom_memory_cache.get(KEY_LOCK) == lock._value

    async def test_acquire_release_key_builder(self, custom_memory_cache, lock):
        async with lock:
            assert await custom_memory_cache.get(KEY_LOCK) is not None
        assert await custom_memory_cache.get(KEY_LOCK) is None

    async def test_release_wrong_token_fails(self, lock):
        await lock.__aenter__()
        lock._value = "random"
        assert await lock.__aexit__("exc_type", "exc_value", "traceback") is None

    async def test_release_wrong_client_fails(self, memory_cache, lock):
        wrong_lock = RedLock(memory_cache, Keys.KEY, 20)
        await lock.__aenter__()
        assert await wrong_lock.__aexit__("exc_type", "exc_value", "traceback") is None

    async def test_float_lease(self, memory_cache):
        lock = RedLock(memory_cache, Keys.KEY, 0.1)
        await lock.__aenter__()
        await asyncio.sleep(0.2)
        assert await lock.__aexit__("exc_type", "exc_value", "traceback") is None


@pytest.mark.redis
class TestRedisRedLock:
    @pytest.fixture
    def lock(self, redis_cache):
        return RedLock(redis_cache, Keys.KEY, 20)

    async def test_acquire_key_builder(self, custom_redis_cache, lock):
        custom_redis_cache.serializer = StringSerializer()
        async with lock:
            assert await custom_redis_cache.get(KEY_LOCK) == lock._value

    async def test_acquire_release_key_builder(self, custom_redis_cache, lock):
        custom_redis_cache.serializer = StringSerializer()
        async with lock:
            assert await custom_redis_cache.get(KEY_LOCK) is not None
        assert await custom_redis_cache.get(KEY_LOCK) is None

    async def test_release_wrong_token_fails(self, lock):
        await lock.__aenter__()
        lock._value = "random"
        assert await lock.__aexit__("exc_type", "exc_value", "traceback") is None

    async def test_release_wrong_client_fails(self, redis_cache, lock):
        wrong_lock = RedLock(redis_cache, Keys.KEY, 20)
        await lock.__aenter__()
        assert await wrong_lock.__aexit__("exc_type", "exc_value", "traceback") is None

    async def test_float_lease(self, redis_cache):
        lock = RedLock(redis_cache, Keys.KEY, 0.1)
        await lock.__aenter__()
        await asyncio.sleep(0.2)
        assert await lock.__aexit__("exc_type", "exc_value", "traceback") is None


@pytest.mark.memcached
class TestMemcachedRedLock:
    @pytest.fixture
    def lock(self, memcached_cache):
        return RedLock(memcached_cache, Keys.KEY, 20)

    async def test_acquire_key_builder(self, custom_memcached_cache, lock):
        custom_memcached_cache.serializer = StringSerializer()
        async with lock:
            assert await custom_memcached_cache.get(KEY_LOCK) == lock._value

    async def test_acquire_release_key_builder(self, custom_memcached_cache, lock):
        custom_memcached_cache.serializer = StringSerializer()
        async with lock:
            assert await custom_memcached_cache.get(KEY_LOCK) is not None
        assert await custom_memcached_cache.get(KEY_LOCK) is None

    async def test_release_wrong_token_succeeds_meh(self, lock):
        await lock.__aenter__()
        lock._value = "random"
        assert await lock.__aexit__("exc_type", "exc_value", "traceback") is None

    async def test_release_wrong_client_succeeds_meh(self, memcached_cache, lock):
        wrong_lock = RedLock(memcached_cache, Keys.KEY, 20)
        await lock.__aenter__()
        assert await wrong_lock.__aexit__("exc_type", "exc_value", "traceback") is None

    async def test_float_lease(self, memcached_cache):
        lock = RedLock(memcached_cache, Keys.KEY, 0.1)
        with pytest.raises(TypeError):
            await lock.__aenter__()


class TestOptimisticLock:
    @pytest.fixture
    def lock(self, cache):
        return OptimisticLock(cache, Keys.KEY)

    async def test_acquire(self, cache, lock):
        await cache.set(Keys.KEY, "value")
        async with lock:
            assert lock._token == await cache._gets(cache._build_key(Keys.KEY))

    async def test_release_does_nothing(self, lock):
        assert await lock.__aexit__("exc_type", "exc_value", "traceback") is None

    async def test_check_and_set_not_existing_never_fails(self, cache, lock):
        async with lock as locked:
            await cache.set(Keys.KEY, "conflicting_value")
            await locked.cas("value")

        assert await cache.get(Keys.KEY) == "value"

    async def test_check_and_set(self, cache, lock):
        await cache.set(Keys.KEY, "previous_value")
        async with lock as locked:
            await locked.cas("value")

        assert await cache.get(Keys.KEY) == "value"

    async def test_check_and_set_fail(self, cache, lock):
        await cache.set(Keys.KEY, "previous_value")
        with pytest.raises(OptimisticLockError):
            async with lock as locked:
                await cache.set(Keys.KEY, "conflicting_value")
                await locked.cas("value")

    async def test_check_and_set_with_int_ttl(self, cache, lock):
        await cache.set(Keys.KEY, "previous_value")
        async with lock as locked:
            await locked.cas("value", ttl=1)

        await asyncio.sleep(1)
        assert await cache.get(Keys.KEY) is None


class TestMemoryOptimisticLock:
    @pytest.fixture
    def lock(self, memory_cache):
        return OptimisticLock(memory_cache, Keys.KEY)

    async def test_acquire_key_builder(self, custom_memory_cache, lock):
        await custom_memory_cache.set(Keys.KEY, "value")
        async with lock:
            assert await custom_memory_cache.get(KEY_LOCK) == lock._token
        await custom_memory_cache.delete(Keys.KEY, "value")

    async def test_check_and_set_with_float_ttl(self, memory_cache, lock):
        await memory_cache.set(Keys.KEY, "previous_value")
        async with lock as locked:
            await locked.cas("value", ttl=0.1)

        await asyncio.sleep(1)
        assert await memory_cache.get(Keys.KEY) is None


@pytest.mark.redis
class TestRedisOptimisticLock:
    @pytest.fixture
    def lock(self, redis_cache):
        return OptimisticLock(redis_cache, Keys.KEY)

    async def test_acquire_key_builder(self, custom_redis_cache, lock):
        custom_redis_cache.serializer = StringSerializer()
        await custom_redis_cache.set(Keys.KEY, "value")
        async with lock:
            assert await custom_redis_cache.get(KEY_LOCK) == lock._token
        await custom_redis_cache.delete(Keys.KEY, "value")

    async def test_check_and_set_with_float_ttl(self, redis_cache, lock):
        await redis_cache.set(Keys.KEY, "previous_value")
        async with lock as locked:
            await locked.cas("value", ttl=0.1)

        await asyncio.sleep(1)
        assert await redis_cache.get(Keys.KEY) is None



import asyncio
from unittest.mock import ANY, MagicMock, create_autospec, patch

import pytest

from aiocache.backends.memory import SimpleMemoryBackend, SimpleMemoryCache
from aiocache.base import BaseCache
from aiocache.serializers import NullSerializer
from ...utils import Keys


@pytest.fixture
def memory(mocker):
    memory = SimpleMemoryBackend()
    mocker.spy(memory, "_cache")
    return memory


class TestSimpleMemoryBackend:
    async def test_get(self, memory):
        await memory._get(Keys.KEY)
        memory._cache.get.assert_called_with(Keys.KEY)

    async def test_gets(self, mocker, memory):
        mocker.spy(memory, "_get")
        await memory._gets(Keys.KEY)
        memory._get.assert_called_with(Keys.KEY, encoding="utf-8", _conn=ANY)

    async def test_set(self, memory):
        await memory._set(Keys.KEY, "value")
        memory._cache.__setitem__.assert_called_with(Keys.KEY, "value")

    async def test_set_no_ttl_no_handle(self, memory):
        await memory._set(Keys.KEY, "value", ttl=0)
        assert Keys.KEY not in memory._handlers

        await memory._set(Keys.KEY, "value")
        assert Keys.KEY not in memory._handlers

    async def test_set_cancel_previous_ttl_handle(self, memory):
        with patch("asyncio.get_running_loop", autospec=True):
            await memory._set(Keys.KEY, "value", ttl=0.1)
            memory._handlers[Keys.KEY].cancel.assert_not_called()

            await memory._set(Keys.KEY, "new_value", ttl=0.1)
            memory._handlers[Keys.KEY].cancel.assert_called_once_with()

    async def test_set_ttl_handle(self, memory):
        await memory._set(Keys.KEY, "value", ttl=100)
        assert Keys.KEY in memory._handlers
        assert isinstance(memory._handlers[Keys.KEY], asyncio.Handle)

    async def test_set_cas_token(self, memory):
        memory._cache.get.return_value = "old_value"
        assert await memory._set(Keys.KEY, "value", _cas_token="old_value") == 1
        memory._cache.__setitem__.assert_called_with(Keys.KEY, "value")

    async def test_set_cas_fail(self, memory):
        memory._cache.get.return_value = "value"
        assert await memory._set(Keys.KEY, "value", _cas_token="old_value") == 0
        assert memory._cache.__setitem__.call_count == 0

    async def test_multi_get(self, memory):
        await memory._multi_get([Keys.KEY, Keys.KEY_1])
        memory._cache.get.assert_any_call(Keys.KEY)
        memory._cache.get.assert_any_call(Keys.KEY_1)

    async def test_multi_set(self, memory):
        await memory._multi_set([(Keys.KEY, "value"), (Keys.KEY_1, "random")])
        memory._cache.__setitem__.assert_any_call(Keys.KEY, "value")
        memory._cache.__setitem__.assert_any_call(Keys.KEY_1, "random")

    async def test_add(self, memory, mocker):
        mocker.spy(memory, "_set")
        await memory._add(Keys.KEY, "value")
        memory._set.assert_called_with(Keys.KEY, "value", ttl=None)

    async def test_add_existing(self, memory):
        memory._cache.__contains__.return_value = True
        with pytest.raises(ValueError):
            await memory._add(Keys.KEY, "value")

    async def test_exists(self, memory):
        await memory._exists(Keys.KEY)
        memory._cache.__contains__.assert_called_with(Keys.KEY)

    async def test_increment(self, memory):
        await memory._increment(Keys.KEY, 2)
        memory._cache.__contains__.assert_called_with(Keys.KEY)
        memory._cache.__setitem__.assert_called_with(Keys.KEY, 2)

    async def test_increment_missing(self, memory):
        memory._cache.__contains__.return_value = True
        memory._cache.__getitem__.return_value = 2
        await memory._increment(Keys.KEY, 2)
        memory._cache.__getitem__.assert_called_with(Keys.KEY)
        memory._cache.__setitem__.assert_called_with(Keys.KEY, 4)

    async def test_increment_typerror(self, memory):
        memory._cache.__contains__.return_value = True
        memory._cache.__getitem__.return_value = "asd"
        with pytest.raises(TypeError):
            await memory._increment(Keys.KEY, 2)

    async def test_expire_no_handle_no_ttl(self, memory):
        memory._cache.__contains__.return_value = True
        await memory._expire(Keys.KEY, 0)
        assert memory._handlers.get(Keys.KEY) is None

    async def test_expire_no_handle_ttl(self, memory):
        memory._cache.__contains__.return_value = True
        await memory._expire(Keys.KEY, 1)
        assert isinstance(memory._handlers.get(Keys.KEY), asyncio.Handle)

    async def test_expire_handle_ttl(self, memory):
        fake = create_autospec(asyncio.TimerHandle, instance=True)
        memory._handlers[Keys.KEY] = fake
        memory._cache.__contains__.return_value = True
        await memory._expire(Keys.KEY, 1)
        assert fake.cancel.call_count == 1
        assert isinstance(memory._handlers.get(Keys.KEY), asyncio.Handle)

    async def test_expire_missing(self, memory):
        memory._cache.__contains__.return_value = False
        assert await memory._expire(Keys.KEY, 1) is False

    async def test_delete(self, memory):
        fake = create_autospec(asyncio.TimerHandle, instance=True)
        memory._handlers[Keys.KEY] = fake
        await memory._delete(Keys.KEY)
        assert fake.cancel.call_count == 1
        assert Keys.KEY not in memory._handlers
        memory._cache.pop.assert_called_with(Keys.KEY, None)

    async def test_delete_missing(self, memory):
        memory._cache.pop.return_value = None
        await memory._delete(Keys.KEY)
        memory._cache.pop.assert_called_with(Keys.KEY, None)

    async def test_delete_non_truthy(self, memory):
        non_truthy = MagicMock(spec_set=("__bool__",))
        non_truthy.__bool__.side_effect = ValueError("Does not implement truthiness")

        with pytest.raises(ValueError):
            bool(non_truthy)

        memory._cache.pop.return_value = non_truthy
        await memory._delete(Keys.KEY)

        assert non_truthy.__bool__.call_count == 1
        memory._cache.pop.assert_called_with(Keys.KEY, None)

    async def test_clear_namespace(self, memory):
        memory._cache.__iter__.return_value = iter(["nma", "nmb", "no"])
        await memory._clear("nm")
        assert memory._cache.pop.call_count == 2
        memory._cache.pop.assert_any_call("nma", None)
        memory._cache.pop.assert_any_call("nmb", None)

    async def test_clear_no_namespace(self, memory):
        memory._handlers = "asdad"
        memory._cache = "asdad"
        await memory._clear()
        memory._handlers = {}
        memory._cache = {}

    async def test_raw(self, memory):
        await memory._raw("get", Keys.KEY)
        memory._cache.get.assert_called_with(Keys.KEY)

        await memory._set(Keys.KEY, "value")
        memory._cache.__setitem__.assert_called_with(Keys.KEY, "value")

    async def test_redlock_release(self, memory):
        memory._cache.get.return_value = "lock"
        assert await memory._redlock_release(Keys.KEY, "lock") == 1
        memory._cache.get.assert_called_with(Keys.KEY)
        memory._cache.pop.assert_called_with(Keys.KEY)

    async def test_redlock_release_nokey(self, memory):
        memory._cache.get.return_value = None
        assert await memory._redlock_release(Keys.KEY, "lock") == 0
        memory._cache.get.assert_called_with(Keys.KEY)
        assert memory._cache.pop.call_count == 0


class TestSimpleMemoryCache:
    def test_name(self):
        assert SimpleMemoryCache.NAME == "memory"

    def test_inheritance(self):
        assert isinstance(SimpleMemoryCache(), BaseCache)

    def test_default_serializer(self):
        assert isinstance(SimpleMemoryCache().serializer, NullSerializer)

    def test_parse_uri_path(self):
        assert SimpleMemoryCache().parse_uri_path("/1/2/3") == {}



import asyncio

from collections import namedtuple

from aiocache import cached, Cache
from aiocache.serializers import PickleSerializer

Result = namedtuple('Result', "content, status")


@cached(
    ttl=10, cache=Cache.REDIS, key="key", serializer=PickleSerializer(),
    port=6379, namespace="main")
async def cached_call():
    return Result("content", 200)


async def test_cached():
    async with Cache(Cache.REDIS, endpoint="127.0.0.1", port=6379, namespace="main") as cache:
        await cached_call()
        exists = await cache.exists("key")
        assert exists is True
        await cache.delete("key")


if __name__ == "__main__":
    asyncio.run(test_cached())
